// Simulator of transitions through place fields.

import javax.swing.*;
import java.util.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.event.*;
import javax.swing.event.*;
import java.awt.image.*;
import java.io.*;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;
import org.xmlpull.v1.XmlPullParserFactory;
import javax.sound.midi.*;



// an influence is a connection from a transition back to
// the place which affects it.  A combination of a place with 
// poss inverter and and/or logic
// can be influenced by a ctrl connector or a place.
class Influence {
    Place m_source = null;
    CtrlConnector m_sourceCtrl = null;

    boolean m_invert = false;
    boolean m_and = false;   

    Influence(Place s, boolean inv, boolean and) {
	m_source = s; m_invert = inv; m_and = and;
    }

    Influence(CtrlConnector s, boolean inv, boolean and) {
	m_sourceCtrl = s; m_invert = inv; m_and = and;
    }
    
    int getSrcX() {
	if (m_source != null) {
	    return m_source.getX();
	}
	return m_sourceCtrl.getCentreX() + m_sourceCtrl.getParentX();
    }

    int getSrcY() {
	if (m_source != null) {
	    return m_source.getY();
	}
	return m_sourceCtrl.getCentreY() + m_sourceCtrl.getParentY();
    }

    boolean enabled() {
	if (m_source != null) {
	    // *** No inversion yet and only OR
	    return m_source.distFromActive == 0;
	}
	else {
	    return m_sourceCtrl.enabled();
	}
    }
}


// a transition is a connection from one place to another.
// it has a probability / weight (which need not add to 1.0)
class Transition {
    Place m_from, m_to;

    CtrlConnector m_toCtrl;

    boolean m_isOnset = false;
    boolean m_isDone  = false;
    double m_prob = 1.0;
    LinkedList<Influence> m_influencedBy = null;   

    static int BLOB_RADIUS = 5;

    transient int x, y, w;

    Transition(Place from, Place to) {
	m_from = from;
	m_to = to;
    }

    Transition(Place from, CtrlConnector to) {
	m_from = from;
	m_to = null;
	m_toCtrl = to;
    }
    

    void setPos(int px, int py, int pw) {
	x = px; y = py; w = pw;
    }

    void setOnset(boolean o) {
	m_isOnset = o;
    }

    void setDone(boolean o) {
	m_isDone = o;
    }

    double getProb() { 
	if (m_influencedBy==null) {
	    return m_prob;
	}
	else {
	    Iterator<Influence> iter = m_influencedBy.iterator();
	    boolean enabled = false;
	    while (iter.hasNext()) {
		Influence inf = iter.next();
		enabled |= inf.enabled();
	    }
	    if (enabled) {
		return m_prob;
	    }
	    else {
		return 0;
	    }
	}
    }

    boolean intersects(int px, int py) {
	int dx = x-px; if (dx<0) dx = -dx;
	int dy = y-py; if (dy<0) dy = -dy;

	if (dx > BLOB_RADIUS || dy > BLOB_RADIUS) return false;
	return true;
    }

    boolean isOnset() { return m_isOnset; }
    boolean isDone() { return m_isDone; }

    boolean isInfluenced() { return m_influencedBy != null; }

    void addInfluence(Influence inf) {
	if (m_influencedBy==null) m_influencedBy = new LinkedList<Influence>();
	m_influencedBy.add(inf);
    }

    void placeRemoved(Place old) {
	// remove any references to old from influenced list
	if (m_influencedBy != null) {
	    Iterator<Influence> iter = m_influencedBy.iterator();
	    while (iter.hasNext()) {
	        Influence inf = iter.next();
		if (inf.m_source == old) {
		    iter.remove();
		}
	    }
	    if (m_influencedBy.size() == 0) {
		m_influencedBy = null;
	    }
	}
    }	
}



/**
 * A thing is a place or a folder or a junction box.
 */
class Thing {
    Folder m_parent = null;
    int x,y; // global coords 

    String m_label = null;

    transient boolean m_sel = false; // selected
    transient boolean m_shifted = false; // shifted bit

    /** 
     * Return true if button intersects this object. 
     * px, py : global pos.
     */
    boolean intersects(int px, int py) { return false; }

    // true if this place lies in the given box
    boolean inBox(int x1, int y1, int x2, int y2) {
	return inBox(x,y,x1,y1,x2-x1, y2-y1);
    }

    void debug() {
	System.out.println("Thing: "+m_parent+" "+x+" "+y+" "+m_label+" "+m_sel+" "+m_shifted);
    }

    void setGlobalX(int gx) {
	x = gx;
    }
    
    void setGlobalY(int gy) {
	y = gy;
    }
    
    int getX() {	
	return x;
    }

    int getY() {
	return y;
    }

    void move(int gx, int gy) {
	shift( gx-x, gy-y );
    }

    // if notify, tell connected things...
    void shift(int dx, int dy) {
	x += dx;
	y += dy;
    }

    void connectorShifted(DataConnector src, int dx, int dy) {	
	if (src.m_xedge==0) dy = 0;
	if (src.m_yedge==0) dx = 0;
	if (dx!=0 || dy != 0) {
	    if (!m_shifted) {
		shift(dx,dy); // shift once...
		m_shifted = true;
	    }
	}
    }

    // return color for link matched to source
    Color getLinkColor() {
	return Color.red;
    }

    // returns true if (px, py) in box
    boolean inBox(int px, int py, int x1, int y1, int w, int h) {
	return (px >= x1 && px <= x1+w && py>=y1 && py<=y1+h);
    }
    
    // returns output data connector.
    // used for mouse interaction of junction / folder border.
    DataConnector getClickedDataConnector() {
	return null;
    }

    // shifted bit used to prevent cycles in movement update
    void clearShifted() {
	m_shifted = false;
    }

    // paint this thing.
    void paint(Graphics gg, MapPanel p, boolean force) {
	return;
    }
}


// A junction box, used for connecting fat links.
// x,y is centre pos.
class Junction extends Thing {
    int m_w = 40; // width of box.
    boolean m_isCtrl = false;

    LinkedList<DataConnector> m_inputs;
    LinkedList<DataConnector> m_outputs;

    Junction() {
	m_inputs  = new LinkedList<DataConnector>();
	m_outputs = new LinkedList<DataConnector>();
    }

    boolean isCtrl() { return m_isCtrl; }

    DataConnector makeInput() {
	DataConnector in;
	if (m_isCtrl) {
	    in = new CtrlConnector("in"+m_inputs.size(), this, -1,0,0,0);
	}
	else {	    
	    in = new DataConnector("in"+m_inputs.size(), this, -1,0,0,0);
	}
	m_inputs.add(in);
	return in;	
    }

    void addInput(DataConnector in) {
	m_inputs.add(in);
    }
    
    void removeOutputs() {
	m_outputs = new LinkedList<DataConnector>();
    }

    DataConnector makeOutput() {
	DataConnector out;
	if (m_isCtrl) {
	    out = new CtrlConnector("out"+m_outputs.size(), this, +1,0,0,0);
	}
	else {	    
	    out = new DataConnector("out"+m_outputs.size(), this, +1,0,0,0);
	}
	m_outputs.add(out);
	return out;	
    }

    void setCtrl() {
	m_w = 10;
	m_isCtrl = true;
    }    
   
    void debug() {
	super.debug();
	System.out.println("  also a Junction: inputs:");
	for (DataConnector conn : m_inputs) { conn.debug(); }
	System.out.println("                   outputs:");
	for (DataConnector conn : m_outputs) { conn.debug(); }
    }
    
    void getSourceFolders(LinkedList<Folder> ret) {
	for (DataConnector in : m_inputs) {
	    in.getSourceFolders(ret);
	}
    }
    
    void getSourceCtrlConnectors(LinkedList<CtrlConnector> ret) {
	for (DataConnector in : m_inputs) {
	    in.getSourceCtrlConnectors(ret);
	}
    }

    // used to link from something...
    DataConnector getClickedDataConnector() {
	// *** Note - only call ONCE! ***
	return makeOutput();
    }

    // get the color from first input.
    Color getLinkColor() {
	DataConnector in = (m_inputs.size()>0) ? m_inputs.get(0) : null;
	if (in != null && in.m_link!=null && in.m_link.m_from !=null)
	    return in.m_link.m_from.getLinkColor();
	return Color.black;
    }

    boolean intersects(int px, int py) {
	return inBox(px, py, x-m_w/2, y-m_w/2, m_w, m_w);
    }

    void shift(int dx, int dy) {
	if (!m_shifted) {
	    x += dx;
	    y += dy;
	    m_shifted = true;

	    for (DataConnector in : m_inputs) {
		in.parentShifted(dx,dy);
	    }
	    for (DataConnector out : m_outputs) { 		    
		out.parentShifted(dx,dy);
	    }
	}
    }

    void connectorShifted(DataConnector src, int dx, int dy) {
	if (src.m_xedge==0) dy = 0;
	if (src.m_yedge==0) dx = 0;
	if (dx!=0 || dy!=0) {
	    if (!m_shifted) {
		x+=dx;
		y+=dy;	    
		
		for (DataConnector in : m_inputs) {
		    if (in != src) { in.parentShifted(dx,dy); }
		}
		for (DataConnector out : m_outputs) {
		    if (out != src) { out.parentShifted(dx,dy); }		    
		}
		m_shifted = true;
	    }
	}
    }

    void deleteConnectorLink(DataConnector dc) {
	if (dc.m_link != null) {
	    dc.m_link.m_parent.removeThing(dc.m_link);
	    dc.m_link.deleteReferences();
	    dc.m_link = null;
	}
    }


    // delete both links connected to this junction.
    void deleteReferences() {
	System.out.println("Deleting references to junction "+this);
	for (DataConnector in : m_inputs) {
	    deleteConnectorLink(in);
	}
	for (DataConnector out : m_outputs) {
	    deleteConnectorLink(out);
	}
	m_inputs  = new LinkedList<DataConnector>();
	m_outputs = new LinkedList<DataConnector>();
    }

    void paint(Graphics gg, MapPanel p, boolean force) {
	if (m_sel) {
	    gg.setColor(new Color(200,220,200,200));
	}
	else {
	    gg.setColor(getLinkColor());
	}
	int x1 = x-m_w/2;
	int y1 = y-m_w/2;
	int x2 = x1 + m_w;
	int y2 = y1 + m_w;

	gg.fillRect(p.getX(x-m_w/2), p.getY(y-m_w/2), p.getD(x1, x2), p.getD(x1,x2));
	return;
    }
}


// A fat arrow from one folder to another.
class DataLink extends Thing {
    DataConnector m_from, m_to;

    int m_w = 20; // 1/2 width of stripe
    int m_aw = 30; // 1/2 width of arrow head
    int m_ad = 10; // depth of arrow.

    boolean m_dragging = false;

    boolean m_isCtrl = false;

    DataLink(DataConnector from, DataConnector to) {
	m_from = from; 
	m_to = to;

	if (m_from instanceof CtrlConnector) {
	    setCtrl();
	}

	// put in backwards links...
	m_from.m_link = this;
	m_to.m_link = this;
    }

    void setCtrl() {
	m_w = 3;
	m_aw = 8;
	m_ad = 8;
	m_isCtrl = true;
    }

    boolean isCtrl() { return m_isCtrl; }

    void debug() {
	System.out.println("  a link "+m_label+" from "+m_from+" to "+m_to);
    }

    // get source folders - other end from end
    void getSourceFolders(LinkedList<Folder> ret, DataConnector end) {
	if (m_from != null && m_from != end) {
	    if (m_from.m_parent instanceof Folder) {
		ret.add((Folder)(m_from.m_parent));
	    }
	    else if (m_from.m_parent instanceof Junction) {
		((Junction)(m_from.m_parent)).getSourceFolders(ret);
	    }
	    else {
		System.out.println("data link other folder problem "+m_from.m_parent);
	    }
	}
    }

    // returns the source ctrl connectors
    void getSourceCtrlConnectors(LinkedList<CtrlConnector> ret) {
	if (m_from != null) {
	    if (m_from.m_parent instanceof Folder) {
		if (m_from instanceof CtrlConnector) {
		    ret.add((CtrlConnector)m_from);
		}
		else {
		    System.out.println("Error - wanted ctrl connector but found "+m_from);
		}
	    }
	    else if (m_from.m_parent instanceof Junction) {
		((Junction)(m_from.m_parent)).getSourceCtrlConnectors(ret);
	    }
	    else {
		System.out.println("data link other folder problem "+m_from.m_parent);
	    }
	}
    }

    void deleteReferencesTo(DataConnector dc) {
	if (dc != null) {
	    dc.linkDeleted();	    
	    dc.m_link = null;
	}
    }

    void deleteReferences() {
	deleteReferencesTo(m_from);
	deleteReferencesTo(m_to);
	m_parent.removeThing(this);
	m_from = null;
	m_to = null;
    }

    // src has been shifted by this amount.
    void shifted(DataConnector src, int dx, int dy) {
	if (m_shifted) return;
	DataConnector other = null;
	if (src == m_to) other = m_from;
	if (src == m_from) other = m_to;
	if (other==null) {
	    System.out.println(" **** Data link shifted error - "+src+" not found !!!");
	    return;
	}

	//	System.out.println("Data link shifted from "+src+" "+dx+" "+dy);
	
	if (src.m_xedge==0) dy = 0;
	if (src.m_yedge==0) dx = 0;
	if (dx!=0 || dy!=0) {
	    //  System.out.println("Data link telling other end link shifted from "+src+" "+dx+" "+dy);
	    m_shifted = true;
	    other.linkShifted(dx, dy);
	}
	else {
	    // System.out.println("Data link swallowed change from "+src+" "+src.m_xedge+" "+src.m_yedge+" "+dx+" "+dy);
	}
    }

    void paintHorizR(Graphics gg, MapPanel p, int x1, int y1, int x2, int y2) {
	gg.fillRect(p.getX(x1), p.getY(y1-m_w), p.getD(x1,x2), p.getD(y1-m_w, y1+m_w));
    }

    void paintHorizL(Graphics gg, MapPanel p, int x1, int y1, int x2, int y2) {
	gg.fillRect(p.getX(x2), p.getY(y1-m_w), p.getD(x2,x1), p.getD(y1-m_w, y1+m_w));
    }

    void paintVertU(Graphics gg, MapPanel p, int x1, int y1, int x2, int y2) {
	gg.fillRect(p.getX(x1-m_w), p.getY(y2), p.getD(x1-m_w, x1+m_w), p.getD(y2, y1));
    }

    void paintVertD(Graphics gg, MapPanel p, int x1, int y1, int x2, int y2) {
	gg.fillRect(p.getX(x1-m_w), p.getY(y1), p.getD(x1-m_w, x1+m_w), p.getD(y1, y2));
    }

    // paint arrow finishing at x,y
    void paintArrowR(Graphics gg, MapPanel p, int x, int y) {
	int xp1 = p.getX(x-m_ad);
	int xp2 = p.getX(x);
	int yptl = p.getY(y-m_aw);
	int ypbl = p.getY(y+m_aw);
	int yptr = p.getY(y-m_w);
	int ypbr = p.getY(y+m_w);
	int[] xp = new int[] {xp1, xp2, xp2, xp1};
	int[] yp = new int[] {yptl, yptr, ypbr, ypbl};
	gg.fillPolygon(xp, yp, 4);
    }

    void paintArrowL(Graphics gg, MapPanel p, int x, int y) {
	int xp1 = p.getX(x);
	int xp2 = p.getX(x+m_ad);
	int yptl = p.getY(y-m_w);
	int ypbl = p.getY(y+m_w);
	int yptr = p.getY(y-m_aw);
	int ypbr = p.getY(y+m_aw);
	int[] xp = new int[] {xp1, xp2, xp2, xp1};
	int[] yp = new int[] {yptl, yptr, ypbr, ypbl};
	gg.fillPolygon(xp, yp, 4);
    }
    
    void paintArrowU(Graphics gg, MapPanel p, int x, int y) {
	int xpbl = p.getX(x-m_aw);
	int xptl = p.getX(x-m_w);
	int xptr = p.getX(x+m_w);
	int xpbr = p.getX(x+m_aw);
	int ypt = p.getY(y);
	int ypb = p.getY(y+m_ad);
	int[] xp = new int[] {xptl, xptr, xpbr, xpbl};
	int[] yp = new int[] {ypt, ypt, ypb, ypb};
	gg.fillPolygon(xp, yp, 4);
    }
    
    void paintArrowD(Graphics gg, MapPanel p, int x, int y) {
	int xpbl = p.getX(x-m_w);
	int xptl = p.getX(x-m_aw);
	int xptr = p.getX(x+m_aw);
	int xpbr = p.getX(x+m_w);
	int ypt = p.getY(y-m_ad);
	int ypb = p.getY(y);
	int[] xp = new int[] {xptl, xptr, xpbr, xpbl};
	int[] yp = new int[] {ypt, ypt, ypb, ypb};
	gg.fillPolygon(xp, yp, 4);
    }

    void setDragging(boolean d) {
	m_dragging = d;
    }
    
    void paint(Graphics gg, MapPanel p, boolean force) {
	if (m_from==null || m_to==null) {
	    System.out.println("error pain data link - from or to null");
	    return;
	}
	
	
	gg.setColor(m_from.getLinkColor()); // new Color(200,200,200,150));
	
	// get position of from dataConnector...
	int x1 = m_from.getConnX() + m_from.m_parent.x;
	int y1 = m_from.getConnY() + m_from.m_parent.y;
	int x2 = m_to.getConnX()+ m_to.m_parent.x ;
	int y2 = m_to.getConnY()+ m_to.m_parent.y ;

	boolean arrow = m_dragging || m_to.m_parent instanceof Folder;
	boolean fromJunc = m_from.m_parent instanceof Junction;
	if (m_from.isRight()) {
	    //	    System.out.println("Right Fat");
	    paintHorizR(gg, p, x1 + (fromJunc ? m_w:0), y1, x2 - (arrow?m_ad:m_w), y2);
	    if (arrow) {
		paintArrowR(gg,p,x2,y1); // use orig Y
	    }
	}
	else if (m_from.isLeft()) {
	    //	    System.out.println("Left Fat");
	    paintHorizL(gg, p, x1 -(fromJunc ? m_w:0) , y1, x2+ (arrow?m_ad:m_w), y2);
	    if (arrow) {
		paintArrowL(gg,p,x2,y1); // use orig Y
	    }
	}
	else if (m_from.isDown()) {
	    //	    System.out.println("Down Fat");
	    paintVertD(gg, p, x1, y1+(fromJunc ? m_w:0), x2, y2 - (arrow?m_ad:m_w));
	    if (arrow) {
		paintArrowD(gg,p,x1,y2); // use orig X
	    }
	}
	else if (m_from.isUp()) {
	    //	    System.out.println("Up Fat");
	    paintVertU(gg, p, x1, y1-(fromJunc ? m_w:0), x2, y2+ (arrow?m_ad:m_w));
	    if (arrow) {
		paintArrowU(gg,p,x1,y2); // orig X
	    }
	}
	else {
	    System.out.println("error - m_from dir unkknown "+m_from);
	}
    }
}


// Connector 
class Connector {
    String m_name;
    Thing m_parent; // part of a folder or a junction box.
    int m_xedge = 0; // -1 / 0 / 1  -- left, top/bottom, right
    int m_yedge = 0; // -1 / 0 / 1  -- top, left/right, bottom
    
    int m_edgeJust  = 0; // -1 / 0 / 1  -- start, med, end
    int m_edgeIndex = 0; // 0 / 1 / 2   -- index of similar things on this part of edge
    //   int m_edgeCount = 1; // number on this edge
    int m_x, m_y, m_width, m_height; // x,y of top left corner.

    transient int m_ax, m_ay; // transient cache of abs pos.

    // positions relative to parent folder.
    void setPos(int x, int y, int width, int height) {
	//System.out.println("conn set pos "+x+" "+y+" "+width+" "+height);
	m_x = x;
	m_y = y;
	m_width = width; 
	m_height = height;
    }
    
    void cacheAbsPosition() {
	m_ax = m_parent.x + m_x;
	m_ay = m_parent.y + m_y;
    }

    // change the parent and make rel to parent pos
    void changeParent(Thing newParent) {
	int ox = m_x + m_parent.x;
	int oy = m_y + m_parent.y;				
	m_parent = newParent;
	m_x = ox -newParent.x;
	m_y = oy -newParent.y;

	//	cacheAbsPosition();
    }


    int getCentreX() { return m_x + m_width/2; }
    int getCentreY() { return m_y + m_height/2; }

    int getParentX() { return m_parent.x; }
    int getParentY() { return m_parent.y; }

    // conn x, y is point on edge for connections.
    int getConnX() { return getCentreX() + (m_xedge*m_width/2); }
    int getConnY() { return getCentreY() + (m_yedge*m_height/2); }

    boolean isLeft()  { return m_xedge == -1; }
    boolean isRight() { return m_xedge == 1; }
    boolean isUp()    { return m_yedge == -1; }
    boolean isDown()  { return m_yedge == 1; }

    // intersects, using relative-to-folder x, y
    boolean intersects( int rx, int ry ) {
	return rx >= m_x && rx <= m_x + m_width && ry >= m_y && ry <= m_y + m_height;
    }
}


// a connector is a place for adding a connection
//
class DataConnector extends Connector {
    DataLink m_link = null;

    // make an opposite connector
    DataConnector makeComplement() {
	DataConnector opp = new DataConnector("tmp", m_parent, -m_xedge, -m_yedge, 0,0);
	opp.m_x = m_x + 100 * m_xedge;
	opp.m_y = m_y + 100 * m_yedge; // fixed size...

	// System.out.println("making an opposite of "+m_name+" "+opp.m_x+" "+opp.m_y+" "+m_x+" "+m_y);
	return opp;
    }

    DataConnector(String name, Thing parent, int xedge, int yedge, int edgeJust, int edgeIndex) {
	m_name = name;
	m_parent = parent;
	m_xedge = xedge;
	m_yedge = yedge;
	m_edgeJust = edgeJust;
	m_edgeIndex = edgeIndex;
    }

    void setDirection(int x1, int y1, int x2, int y2) {
	if (x1==x2) m_xedge=0; else if (x1>x2) m_xedge = -1; else m_xedge = 1;
	if (y1==y2) m_yedge=0; else if (y1>y2) m_yedge = -1; else m_yedge = 1;
    }

    void getSourceFolders(LinkedList<Folder> ret) {
	if (m_link!=null) m_link.getSourceFolders(ret, this);
    }

    Junction getSourceJunction() {
	if (m_link!=null) {
	    if (m_link.m_from != null) {
		if (m_link.m_from.m_parent instanceof Junction) {
		    return (Junction) (m_link.m_from.m_parent);
		}
	    }
	}
	return null;
    }

    void getSourceCtrlConnectors(LinkedList<CtrlConnector> ret) {
	if (m_link!=null) m_link.getSourceCtrlConnectors(ret);
    }

    void debug() {
	System.out.println("   data connector "+m_name+" parent "+m_parent);
	if (m_link!=null) {
	    m_link.debug();
	}
    }
    
    void cacheAbsPosition() {
	m_ax = m_parent.x + getConnX();
	m_ay = m_parent.y + getConnY();
	if (m_link != null) {
	    System.out.println("cache abs pos "+m_ax+" "+m_ay);
	}
    }

    Color getLinkColor() {
	return m_parent.getLinkColor();
    }

    // this connection shifted by this amount (parent moved).
    void parentShifted(int dx, int dy) {
	if (m_link != null) {
	    // System.out.println("data connector "+m_name+" parent shifted, telling link ");
	    m_link.shifted(this, dx, dy);
	}
    }

    // positions relative to parent folder.
    void setPos(int x, int y, int width, int height) {
	// do we tell connection we've moved?
	int oax = m_ax;
	int oay = m_ay;
	m_x = x;
	m_y = y;
	m_width = width; 
	m_height = height;

	cacheAbsPosition();
	int dx = m_ax - oax;
	int dy = m_ay - oay;
	if (m_link != null) {
	    System.out.println("setpos telling link pos has moved " +dx+" "+dy+" "+x+" "+y);
	    m_link.shifted(this, dx, dy);
	}

    }
    
    // the link has been shifted by a bit...
    void linkShifted(int dx, int dy) {
	System.out.println(" connector "+m_name+" "+this+" telling parent "+m_parent+" that link shifted "+dx+" "+dy);
	if (m_xedge==0 && dx!=0) {
	    m_parent.connectorShifted(this, dx,dy);
	}
	else if (m_yedge==0 && dy!=0) {
	    m_parent.connectorShifted(this, dx,dy);
	}
	else {
	    System.out.println("connector "+this+" swallowing link shift "+dx+" "+dy +" xed "+m_xedge+" yed "+m_yedge);
	}
    }
    
    void linkDeleted() {
	m_link = null;
    }

}


// a control connector is for ctrl signals.
class CtrlConnector extends DataConnector {
    // influence infl ???

    boolean m_isClock  = false;
    boolean m_isReset  = false;
    boolean m_isMatch    = false;

    static Color m_rcol = new Color(255,0,0,100);
    static Color m_ccol = new Color(0,255,0,100);
    static Color m_fcol = new Color(0,0,255,100);

    LinkedList<Transition> links_in = null;

    // returns a color if special type of ctrl link.
    Color getSpecialColor() {
	if (isReset()) { return m_rcol; }
	if (isClock()) { return m_ccol; }
	if (isMatch()) { return m_fcol; }
	return null;
    }

    String getType() {
	if (isReset()) return "reset";
	if (isClock()) return "clock";
	if (isMatch())   return "match";
	return "normal";
    }

    Color getLinkColor() {
	Color sc = getSpecialColor();
	return (sc!=null) ? sc : m_parent.getLinkColor();
    }

    DataConnector makeComplement() {
	CtrlConnector opp = new CtrlConnector("tmp", m_parent, -m_xedge, -m_yedge, 0,0);
	opp.m_x = m_x + 100 * m_xedge;
	opp.m_y = m_y + 100 * m_yedge; // fixed size...
	opp.m_isClock = m_isClock;
	opp.m_isReset = m_isReset;
	opp.m_isMatch   = m_isMatch;

	return opp;
    }

    void addIncoming(Transition from) {
	if (links_in==null) {
	    links_in= new LinkedList<Transition>();
	}
	links_in.add(from);
    }

    // Check whether this ctrl connection is enabled.
    boolean enabled() {
	LinkedList<CtrlConnector> sources = new LinkedList<CtrlConnector>();
	getSourceCtrlConnectors(sources);
	
	//	System.out.println("checking enabled for "+getType()+" sources size is "+sources.size());

	if (sources.size()==0) {
	    if (isClock()) return true;  
	    else if (isReset()) return false;  
	    else if (isMatch()) return false;  
	    else if (m_parent instanceof Folder) {
		return ((Folder)m_parent).getDefaultCtrlOutput();
		//return false; // or whatever default is.
	    }
	    else {
		// *** Space for things like "output enable"
		return false;
	    }
	}
	
	boolean val = false;
	for (CtrlConnector src : sources) {
	    // Do an -OR- of any places attached...
	    if (src.links_in != null) {
		for (Iterator<Transition> iter = src.links_in.iterator(); iter.hasNext();) {
		    val |= iter.next().m_from.distFromActive == 0;
		}
	    }
	    else {
		val |= ((Folder)src.m_parent).getDefaultCtrlOutput();
	    }
	}
	return val;
    }

    boolean isReset() { return m_isReset; }
    boolean isClock() { return m_isClock; }
    boolean isMatch()   { return m_isMatch; }

    CtrlConnector(String name, Thing parent, int xedge, int yedge, int edgeJust, int edgeIndex) {
	super(name, parent, xedge, yedge, edgeJust, edgeIndex);
    }
}



/**
 * A folder - a logical grouping of places / folders.
 * x,y pos is below title.
 * [title]__________________
 * /                        |
 * |                        |
 * |                        |
 * |                        |
 * |                        |
 * |_______________________/
 *
 *
 *
 * Now adding connectors:
 *
 */
class Folder extends Thing {
    int m_width=200, m_height=160; // width / height.

    int m_titleHeight = 30; // size of title bar...
    int m_titleWidth  = 150;
    int m_resizeWidth = 30; // size of resize triangle

    int m_connectorWidth = 12; // width of connector bars

    // data & control connectors.
    ArrayList<DataConnector> m_dconns = null;
    
    DataConnector m_t1;
    DataConnector m_t2;
    DataConnector m_l1;
    DataConnector m_l2;
    DataConnector m_b1;
    DataConnector m_b2;
    DataConnector m_r1;
    DataConnector m_r2;

    CtrlConnector m_tl;
    CtrlConnector m_tr;
    CtrlConnector m_bl;
    CtrlConnector m_br;
    CtrlConnector m_tm;
    CtrlConnector m_lm;
    CtrlConnector m_rm;
    CtrlConnector m_bm;
    CtrlConnector m_match;
    CtrlConnector m_clk;
    CtrlConnector m_reset;

    ArrayList<Thing> m_things = new ArrayList<Thing>();

    // visible transitions (with blobs) owned by this folder.
    LinkedList<Transition> m_visibleTransitions = new LinkedList<Transition>();

    // hints for mouse interaction
    DataConnector m_clickedData;

    // interface for complex behaviours.
    Behaviour m_behaviour = null;

    public Folder() {
	this("noname", 0,0);
    }

    // used for mouse interaction of junction / folder border.
    DataConnector getClickedDataConnector() {
	return m_clickedData;
    }
    
    void debug() {
	super.debug();
	System.out.println("  also a Folder: "+m_parent+" clicked data "+m_clickedData);
    }

    public Folder(String name, int x1, int y1) {
	x = x1;
	y = y1;
	m_label = name;
       
	// 8 poss data connectors.
	m_dconns = new ArrayList<DataConnector>();
	m_t1 = new DataConnector("T1", this, 0, -1, -1, 0);
	m_t2 = new DataConnector("T2", this, 0, -1, +1, 0);
	m_l1 = new DataConnector("L1", this, -1, 0, -1, 0);
	m_l2 = new DataConnector("L2", this, -1, 0, +1, 0);
	m_b1 = new DataConnector("B1", this, 0, 1, -1, 0);
	m_b2 = new DataConnector("B2", this, 0, 1, +1, 0);
	m_r1 = new DataConnector("R1", this, 1, 0, -1, 0);
	m_r2 = new DataConnector("R2", this, 1, 0, +1, 0);

	m_dconns.add(m_t1); 
	m_dconns.add(m_t2);
	m_dconns.add(m_l1);
	m_dconns.add(m_l2);
	m_dconns.add(m_b1);
	m_dconns.add(m_b2);
	m_dconns.add(m_r1);
	m_dconns.add(m_r2);


	// corners
	/*
	m_tl = new CtrlConnector("TL", this, -1, -1, 0, 0);
	m_tr = new CtrlConnector("TR", this,  1, -1, 0, 0);
	m_bl = new CtrlConnector("BL", this, -1,  1, 0, 0);
	m_br = new CtrlConnector("BR", this,  1,  1, 0, 0);
	*/

	// midpoints
	m_tm = new CtrlConnector("TM", this,  0, -1, 0, 0);
	m_lm = new CtrlConnector("LM", this, -1,  0, 0, 0);
	m_rm = new CtrlConnector("RM", this,  1,  0, 0, 0);
	m_bm = new CtrlConnector("BM", this,  0,  1, 0, 0);

	// clk and reset
	m_clk  = new CtrlConnector("Clk", this,  -1, 0, -1, 0);
	m_clk.m_isClock = true;
	m_reset = new CtrlConnector("Reset", this,-1, 0, -1, 1);
	m_reset.m_isReset = true;
	m_match = new CtrlConnector("Match", this,+1, 0, -1, 2);
	m_match.m_isMatch = true;

	/*
	m_dconns.add( m_tl );
	m_dconns.add( m_tr );
	m_dconns.add( m_bl );
	m_dconns.add( m_br );
	*/

	m_dconns.add( m_tm );
	m_dconns.add( m_lm );
	m_dconns.add( m_rm );
	m_dconns.add( m_bm );

	m_dconns.add( m_clk );
	m_dconns.add( m_reset );
	m_dconns.add( m_match );

	layoutConnectors();
    }
    
    
    void setBehaviour(Behaviour b) {
	m_behaviour = b;
    }

    // returns true if folder wants to activate its default ctrl outputs.
    // returns false if any predictions weren't found in inputs.
    // returns true if inputs are superset of predictions, or no prediction.
    boolean getDefaultCtrlOutput() {
	System.out.println("Folder "+m_label+" asked for default ctrl output signal.");
	
	HashMap<String,String> predOutputs = new HashMap<String,String>();
	// Build a hash of predicted outputs.
	Iterator<Thing> mythings = m_things.iterator();
	while (mythings.hasNext()) {
	    // check whether 
	    Thing t = mythings.next();
	    if (t instanceof Place) {
		Place p = (Place) t;
		if (p.distFromActive ==0 && p.m_outputCmd != null) {
		    if (p.m_label != null) {
			predOutputs.put(p.m_label, "pred");
		    }
		    else {
			predOutputs.put("note-"+(p.y-p.m_parent.y), "pred");
		    }
		}
	    }
	}
		
	// Compare of inputs with predicted outputs...
	LinkedList<Folder> sources = getInputFolders();
	for (Folder src : sources) {
	    Iterator<Thing> iter = src.m_things.iterator();
	    if (iter != null) {
		while (iter.hasNext()) {		
		    Thing t = iter.next();
		    if (t instanceof Place) {
			Place p = (Place)t;

			// base inputs on next dist from active...
			if (p.nextDistFromActive == 0 && p.m_outputCmd != null) {
			    // we have activity!  Make a new place!!	
			    if (p.m_label != null) {
				System.out.println("Found activity "+p.m_label);			

				predOutputs.put(p.m_label, "found");
			    }
			    else {
				predOutputs.put("note-"+(p.y-p.m_parent.y), "found");
			    }
			}		   
		    }
		}
	    }
	    else {
		System.out.println("rec clocked by couldn't find input");
	    }
	}

	// Go through hash seeing if any preds weren't found.
	Iterator<String> keys = predOutputs.keySet().iterator();
	while (keys.hasNext()) {
	    String k = keys.next();
	    if (predOutputs.get(k).equals("pred")) {
		System.out.println("Prediction "+k+" not found in input");
		return false;
	    }
	}

	System.out.println("All as predicted");
	return true;
    }


    void preResize() {
	// called before a resize - get the connectors to compute abs position.
	for (int c=0; c<m_dconns.size(); c++) {
	    m_dconns.get(c).cacheAbsPosition();
	}
    }    

    // Call this on resize.
    void layoutConnectors() {
	int xp1 = -m_connectorWidth;
	int xp2 = 0;
	int xp3 = m_width/2;
	int xp4 = m_width;

	int yp1 = -m_connectorWidth;
	int yp2 = 0;
	int yp3 = m_height/2;
	int yp4 = m_height;

	int cw = m_connectorWidth;
	int cw2 = m_connectorWidth / 2;
	int w2  = m_width / 2;
	int h2  = m_height / 2;

	// data connectors : Top
	m_t1.setPos(0, yp1, w2-cw2, cw);
	m_t2.setPos(xp3+cw2, yp1, w2-cw2, cw);

	// Bottom
	m_b1.setPos(0, yp4, w2-cw2, cw);
	m_b2.setPos(xp3+cw2, yp4, w2-cw2, cw);

	// Left
	m_l1.setPos(-cw, yp2, cw, h2-cw2);
	m_l2.setPos(-cw, yp3+cw2, cw, h2-cw2  -cw-cw); // adj height for clk / reset / fix

	// Right
	m_r1.setPos(xp4, yp2, cw, h2-cw2);
	m_r2.setPos(xp4, yp3+cw2, cw, h2-cw2  -cw);


	// --- ctrl connectors...
	/*
	m_tl.setPos(xp1,yp1,cw,cw);
	m_bl.setPos(xp1,yp4,cw,cw);
	m_tr.setPos(xp4,yp1,cw,cw);
	m_br.setPos(xp4,yp4,cw,cw);
	*/


	// midpoint ones...
	m_tm.setPos(xp3-cw2, yp1, cw, cw);
	m_bm.setPos(xp3-cw2, yp4, cw, cw);
	m_lm.setPos(xp2-cw, yp3-cw2, cw, cw);
	m_rm.setPos(xp4, yp3-cw2, cw, cw);

	// clk and reset
	m_clk.setPos(xp1, yp4-cw, cw, cw);
	m_reset.setPos(xp1, yp4-cw-cw, cw, cw);

	// match
	m_match.setPos(xp4, yp4-cw, cw, cw);
    }

    
    // copy the input connections from orig folder, making extra junctions as needed
    void copyConnectionsFrom(Folder orig, int dx, int dy) {
	for (int c=0; c<orig.m_dconns.size(); c++) {	
	    DataConnector oc = orig.m_dconns.get(c);
	    if (oc.m_link != null) {
		DataConnector fc = oc.m_link.m_from;
		if (fc != null && fc != oc) {
		    if (fc.m_parent instanceof Junction) {
			Junction oldj = (Junction) fc.m_parent;

			// OK! we've found an original junction.
			// Make a new junction.
			Junction newj = new Junction();
			m_parent.addThing(newj);

			newj.x = oldj.x + dx;
			newj.y = oldj.y + dy;
			newj.m_w = oldj.m_w;
			newj.m_isCtrl = oldj.m_isCtrl;

			DataConnector oldjout = oldj.makeOutput();
			oldjout.setDirection(oldj.x, oldj.y, newj.x, newj.y);
			
			DataConnector newjin  = newj.makeInput();
			newjin.setDirection(newj.x, newj.y, oldj.x, oldj.y);

			DataConnector newjout = newj.makeOutput();
			DataConnector fin = m_dconns.get(c);
			newjout.m_xedge = -fin.m_xedge;
			newjout.m_yedge = -fin.m_yedge;
			
      			// Make a link from the oldjunction to the new one.
			DataLink jlink = new DataLink(oldjout, newjin);
			if (oldjout instanceof CtrlConnector) jlink.setCtrl();
			m_parent.addThing(jlink);

			// make a link from new junction to this connector
			DataLink flink = new DataLink(newjout, m_dconns.get(c));
			if (oldjout instanceof CtrlConnector) flink.setCtrl();
			m_parent.addThing(flink);
		    }
		}		    
	    }
	}	
    }


    void clearShifted() {
	m_shifted = false; 

	Iterator<Thing> iter = m_things.iterator();
	while (iter.hasNext()) {
	    Thing t = iter.next();
	    t.clearShifted();
	}
    }

    void shift(int dx, int dy) {
	if (!m_shifted) {
	    super.shift(dx, dy);
	    for (int i=0; i<m_things.size(); i++) {
		Thing p = m_things.get(i);
		p.shift(dx,dy);	    
	    }
	    m_shifted = true;
	}
    }
    
    // Move this folder + fat links
    void move(int px, int py) {
	int dx = px-x;
	int dy = py-y;
	shift(dx,dy);
	shiftLinks(dx,dy);
    }
    
    // tell the connectors they've shifted
    void shiftLinks(int dx, int dy) {
	for (int c=0; c<m_dconns.size(); c++) {
	    m_dconns.get(c).parentShifted(dx,dy);
	}
    }

    // tell the connectors they've shifted
    void shiftLinksExcept(int dx, int dy, DataConnector src) {
	for (int c=0; c<m_dconns.size(); c++) {
	    if (m_dconns.get(c) != src) {
		m_dconns.get(c).parentShifted(dx,dy);
	    }
	}
    }

    void connectorShifted(DataConnector src, int dx, int dy) {	
	if (m_shifted) return;
	if (src.m_xedge==0) dy = 0;
	if (src.m_yedge==0) dx = 0;
	if (dx!=0 || dy != 0) {
	    // System.out.println(this+" conn shifted "+dx+" "+dy);
	    shift(dx,dy); // shift once...
	    shiftLinksExcept(dx, dy, src);
	    m_shifted = true; 
	}
    }
    
    public String toString() { return m_label+" "+x+" "+y; }

    static int maxColors = 20;

    public static float getHue(String tag) { 
	if (tag==null) return 0f;
	float hash = (tag.hashCode() * 17) % maxColors;
	float hue = hash / (maxColors-1);
	return hue; 
    }

    // get color based on title with given opaque value
    Color getColor(int a) {
	Color col = Color.getHSBColor(getHue(m_label), 0.3f, 0.8f);
	Color ret = new Color(col.getRed(), col.getGreen(), col.getBlue(), a);
	return ret;
    }   

    // return color for link matched to source
    Color getLinkColor() {
	return getColor( 150 );
    }

    // paint this thing.
    void paint(Graphics gg, MapPanel p, boolean force) {

	if (force) {
	// draw a rectangle...
	Color col = getColor(m_sel ? 150 : 100);
	Color col2 = getColor(m_sel ? 120 : 75);
	Color bg2 = new Color(0,0,0,50);
	Color bg3 = new Color(0,0,0,60);

	Color dcol = getColor( m_sel ? 105 : 35);
	Color ccol = getColor( m_sel ? 130 : 105);

	m_titleWidth = m_width + m_connectorWidth +  m_connectorWidth;
	gg.setColor(col2);	

	gg.fillRect(p.getX(x-m_connectorWidth), p.getY(y-m_titleHeight-m_connectorWidth), p.getD(x-m_connectorWidth, x-m_connectorWidth+m_titleWidth), p.getD(y-m_titleHeight-m_connectorWidth, y-m_titleHeight-m_connectorWidth+m_titleHeight));

	if (m_behaviour==null) {
	    gg.setColor(Color.yellow);
	    gg.drawString(m_label, p.getX(x+4), p.getY(y-4-m_connectorWidth));
	}
	else {
	    gg.setColor(Color.green);
	    gg.drawString(m_label + " >>> " + m_behaviour.getClass().getName(), p.getX(x+4), p.getY(y-4-m_connectorWidth));
	}

	gg.setColor(col);	
	gg.fillRect(p.getX(x), p.getY(y), p.getE(x,m_width), p.getE(y,m_height));

	// a line...
	gg.setColor(bg2);	
	gg.drawRect(p.getX(x-m_connectorWidth)-1, p.getY(y-m_connectorWidth)-1, p.getD(x-m_connectorWidth, x-m_connectorWidth+m_titleWidth)+2, p.getD(y-m_connectorWidth, y-m_connectorWidth+m_height+m_connectorWidth+m_connectorWidth)+2);

	// draw resize triangle BR
	int xp1 = p.getX(x+m_width-m_resizeWidth);
	int xp2 = p.getX(x+m_width);
	int yp1 = p.getY(y+m_height-m_resizeWidth);
	int yp2 = p.getY(y+m_height);

	int[] xp = new int[] {xp1, xp2, xp2};
	int[] yp = new int[] {yp2, yp1, yp2};
	gg.setColor(col2);	
	gg.fillPolygon(xp, yp, 3);

	// draw resize triangle TL
	xp1 = p.getX(x);
	xp2 = p.getX(x+m_resizeWidth);
	yp1 = p.getY(y);
	yp2 = p.getY(y+m_resizeWidth);
	xp = new int[] {xp1, xp2, xp1};
	yp = new int[] {yp1, yp1, yp2};       
	gg.fillPolygon(xp, yp, 3);

	// draw the connectors...
	gg.setColor(dcol);
	for (int i=0; i<m_dconns.size(); i++) {
	    DataConnector dc = m_dconns.get(i);
	    if (dc instanceof CtrlConnector) {
		CtrlConnector cc = (CtrlConnector)dc;
		Color sc = cc.getSpecialColor();
		if (sc!=null) { 
		    gg.setColor(sc); 
		}
		else {
		    gg.setColor(ccol);
		}
		gg.fillRect(p.getX(x + dc.m_x), p.getY(y+ dc.m_y), p.getE(x + dc.m_x, dc.m_width), p.getE(y+ dc.m_y, dc.m_height));
	    }
	    else {
		gg.fillRect(p.getX(x + dc.m_x), p.getY(y+ dc.m_y), p.getD(x + dc.m_x, x + dc.m_x+dc.m_width), p.getD(y+ dc.m_y, y+ dc.m_y+dc.m_height));
	    }
	}
	}
	paintChildren(gg, p, force);
    }

    void paintChildren(Graphics gg, MapPanel pan, boolean force) {
	int i;
	// draw children... (if not collapsed TBD)
	for (i=0; i<m_things.size(); i++) {
	    Thing p = m_things.get(i);
	    p.paint(gg, pan, force);	    
	}
    }

    void addThing(Thing t) {
	m_things.add(t);
	t.m_parent = this;
    }

    void removeThing(DataLink l) {
	if (m_things.remove(l)) {
	    System.out.println("removed data link ok");
	}
    }
    
    // return source folders (could be cached)
    LinkedList<Folder> getInputFolders() {
	LinkedList<Folder> ret = new LinkedList<Folder>();
	for (int i=0; i<m_dconns.size(); i++) {
	    DataConnector dc = m_dconns.get(i);
	    
	    if (!(dc instanceof CtrlConnector)) {
		if (dc.m_link != null) {
		    // *** should check that it's not this one...
		    // as we're getting output folders too...
		    dc.getSourceFolders(ret);
		}
	    }
	}
	return ret;
    }

    // get precursor junctions... used for recorder.fix
    LinkedList<Junction> getInputJunctions() {
	LinkedList<Junction> ret = new LinkedList<Junction>();
	for (int i=0; i<m_dconns.size(); i++) {
	    DataConnector dc = m_dconns.get(i);
	    	    
	    if (dc.m_link != null) {
		Junction j = dc.getSourceJunction();
		if (j!=null) {
		    ret.add(j);
		}
	    }
	}
	return ret;
    }

    // make sure it's big enough...
    static int minBorder = 16;

    // make sure big enough to contain t.
    void expandToFit(Thing t) {
	int tx = t.x;
	int ty = t.y;
	
	System.out.println("exp to fit "+x+" "+y+" "+t.x+" "+t.y);

	// expand top left.
	if (tx-minBorder < x) {
	    int xinc = x-tx;
	    x = tx-minBorder;
	    m_width += xinc;
	}
	if (ty-minBorder-m_titleHeight < y) {
	    int yinc = y-ty+minBorder+m_titleHeight;
	    y = ty-minBorder-m_titleHeight;
	    m_height += yinc;
	}

	// expand bottom right
	int tw = minBorder;
	int th = minBorder;
	if (t instanceof Folder) {
	    Folder ft = (Folder)t;
	    tw = ft.m_width + minBorder;
	    th = ft.m_height + minBorder;
	}

	int rx = tx + tw;
	int ry = ty + th;
	int rx2 = x + m_width;
	int ry2 = y + m_height;
	    
	if (rx > rx2) {
	    int inc = rx - rx2;
	    m_width += inc;
	}
	if (ry > ry2) {
	    int inc = ry - ry2;
	    m_height += inc;
	}
    }

    int folderNum = 0;

    // group the selected items and put them in a folder...
    // if f is null, put in this folder.
    Folder groupSelected(Folder f) {
	Iterator<Thing> iter = m_things.iterator();
	Folder newFolder = null;
	while (iter.hasNext()) {
	    Thing t = iter.next();
	    if (t.m_sel && t != f) {
		int tx = t.getX();
		int ty = t.getY();
		if (f == null) {	
		    f = new Folder("Group "+folderNum, 100, 100);
		    folderNum++;
		    newFolder = f;
		    // set initial pos of folder
		    f.setGlobalX(t.getX() - 16);
		    f.setGlobalY(t.getY() - 16);
		    f.m_width = 60;
		    f.m_height = 60;
		}
		f.addThing(t);
		iter.remove();
		f.expandToFit(t);
		t.m_sel = false; // deselect.
	    }

	    if (t instanceof Folder && t != f) {
		f = ((Folder)t).groupSelected(f);
	    }
	}	

	if (newFolder != null) {
	    newFolder.layoutConnectors();
	    addThing(newFolder);
	}
	return f;
    }

    // ungroup any selected folders (recursively)
    void ungroupSelected() {
	Iterator<Thing> iter = m_things.iterator();
	LinkedList<Thing> newThings = new LinkedList<Thing>();
	while (iter.hasNext()) {
	    Thing t = iter.next();

	    if (t instanceof Folder) {
		Folder f = (Folder)t;
		if (t.m_sel) {
		    iter.remove();
		    // add contents here...
		    Iterator<Thing> iter2 = f.m_things.iterator();
		    while (iter2.hasNext()) {			
			newThings.add(iter2.next());
		    }
		}
		else {
		    ((Folder)t).ungroupSelected();
		}
	    }
	}	

	// actually add all the collapsed things.
	Iterator<Thing> iter2 = newThings.iterator();
	while (iter2.hasNext()) {	    
	    addThing(iter2.next());
	}
    }


    void selectOnly(Thing st) {
	Iterator<Thing> iter = m_things.iterator();
	while (iter.hasNext()) {
	    Thing t = iter.next();
	    if (t == st) {
		t.m_sel = !t.m_sel;
		t.debug();
	    }
	    else {
		t.m_sel = false;
	    }

	    if (t instanceof Folder) {
		// recurse
		((Folder)t).selectOnly(st);
	    }
	}	
    }

    // delete seleced children
    // if selAll, delete all
    // deleted is a list of deleted things.
    void deleteSelected(LinkedList<Thing> deleted, boolean selAll) {
	ArrayList<Thing> origThings = new ArrayList<Thing>(m_things);
	Iterator<Thing> iter = origThings.iterator();

	while (iter.hasNext()) {
	    Thing t = iter.next();

	    if (t.m_sel || selAll) {
		deleted.add(t);
		// iter.remove();		
		if (t instanceof Junction) {
		    ((Junction)t).deleteReferences();
		}
	    }
	    
	    if (t instanceof Folder) {
		// recurse, just in case.
		((Folder)t).deleteSelected(deleted, t.m_sel || selAll);
	    }
	    else if (t instanceof DataLink) {
		// ... not handled yet...
	    }
	}

	// actually remove from Things....	
	iter = deleted.iterator();
	while (iter.hasNext()) {
	    m_things.remove(iter.next());
	}	
    }
    
    // delete references to all items in deleted list
    void deleteReferencesTo(LinkedList<Thing> deleted) {
	Iterator<Thing> iter = m_things.iterator();
	while (iter.hasNext()) {
	    Thing t = iter.next();
	    if (t instanceof Place) {
		Iterator<Thing> del = deleted.iterator();
		while (del.hasNext()) {
		    Thing dt = del.next();
		    if (dt instanceof Place) {
			((Place)t).deleteReferencesTo((Place)dt);
		    }
		}
	    }
	    else if (t instanceof Folder) {
		((Folder)t).deleteReferencesTo(deleted);
	    }
	    //	    else if (t instanceof Junction) {
	    //		((Junction)t).deleteReferences();
	    // }
	}
    }

    // true if this place lies in the given box
    // coords relative to parent
    boolean inBox(int x1, int y1, int x2, int y2) {
	// *** TBD : fix this (just checking TL corner at present)
	return inBox(x,y,x1,y1,x2-x1,y2-y1);
    }

    // Select all with given box (in local coords)
    void selectBox(int x1, int y1, int x2, int y2) {
	Iterator<Thing> iter = m_things.iterator();
	while (iter.hasNext()) {
	    Thing t = iter.next();	    
	    t.m_sel = t.inBox(x1,y1,x2,y2);
	    if (t instanceof Folder && !t.m_sel) {
		((Folder)t).selectBox(x1,y1,x2,y2);
	    }
	}
    }

    // add selected to list.
    // if folder selected, add it.
    void addToSelectedThings(LinkedList<Thing> selThings) {
	Iterator<Thing> iter = m_things.iterator();
	while (iter.hasNext()) {
	    Thing t = iter.next();
	    if (t.m_sel) {
		selThings.add(t);
	    }
	    else if (t instanceof Folder) {
		((Folder)t).addToSelectedThings(selThings);
	    }
	}
    }

    // look for 1st transition which intersects x,y
    // TBD : multi-level transitions.
    Transition findTransition(int x, int y) {
	Iterator<Transition> iter = m_visibleTransitions.iterator();
	while (iter.hasNext()) {
	    Transition trans = iter.next();
	    if (trans.intersects(x,y)) {
		return trans;
	    }
	}
	return null;
    }

    // what we hit...
    String m_clickHint = "";
    

    // look for folder contents at px, py
    Folder findContext(int px, int py) {
	for (int i=0; i<m_things.size(); i++) {
	    Thing p = m_things.get(i);
	    if (p instanceof Folder) {
		Folder f = (Folder) p;
		if (f.intersects(px, py)) {
		    // could be title / contents / resize/
		    if (f.clickedContents()) {
			System.out.println("found contents of "+f);
			// try child first.
			Folder grandchild = f.findContext(px, py);
			if (grandchild != null) {
			    // some child of folder.
			    return grandchild;
			}			
			else {
			    return f;
			}
		    }
		    else {
			// clicked on title or resize of folder.
			return f;
		    }
		}
	    }
	}	
	return this;
    }


    
    // look for 1st place which intersects px,py
    // px, py global.
    Thing findThing(int px, int py) {
	for (int i=0; i<m_things.size(); i++) {
	    Thing p = m_things.get(i);
	    if (p instanceof Place) {
		if (p.intersects(px,py)) {		
		    return p;
		}		
	    }
	    else if (p instanceof Junction) {
		if (p.intersects(px,py)) {		
		    return p;
		}				
	    }
	    else if (p instanceof Folder) {
		Folder f = (Folder) p;
		if (f.intersects(px, py)) {
		    // could be title / contents / resize/
		    if (f.clickedContents()) {
			System.out.println("found contents of "+f);
			// try child first.
			Thing grandchild = f.findThing(px, py);
			if (grandchild != null) {
			    // some child of folder.
			    return grandchild;
			}			
			else {
			    // clicked in empty space.
			    // don't select the folder here!
			    ; // return null;
			}
		    }
		    else {
			// clicked on title or resize of folder.
			return f;
		    }
		}
		else {
		    // recurse anyway just in case...
		    Thing grandchild = f.findThing(px, py);
		    if (grandchild != null) {
			// some child of folder.
			return grandchild;
		    }			
		}
	    }
	}	
	return null;
    }
    
    // returns boolean intersects w bbox
    boolean intersects(int px, int py) {
	m_clickHint="";
	
	// still excluding bounding box.
	if (!inBox( px, py, x-m_connectorWidth, y-m_titleHeight-m_connectorWidth, m_width+m_connectorWidth+m_connectorWidth, 
		    m_height+m_titleHeight+m_connectorWidth+m_connectorWidth)) {
	    return false;
	}
	
	// Does it intersect title bar???
	if (inBox(px, py, x, y-m_titleHeight-m_connectorWidth, m_titleWidth, m_titleHeight)) {
	    System.out.println("clicked title");
	    m_clickHint = "title";
	    return true;
	}

	// Does it intersect resize box???
	if (inBox(px, py, x+m_width-m_resizeWidth, y+m_height-m_resizeWidth, m_resizeWidth, m_resizeWidth)) {
	    System.out.println("clicked resize");
	    m_clickHint = "resize";
	    return true;
	}

	if (inBox(px, py, x, y, m_resizeWidth, m_resizeWidth)) {
	    System.out.println("clicked resize tl");
	    m_clickHint = "resizetl";
	    return true;
	}

	// check for contents
	if (inBox( px, py, x, y, m_width, m_height)) {
	    m_clickHint = "contents";
	    return true;
	}


	for (int j=0; j<m_dconns.size(); j++) {
	    DataConnector dc = m_dconns.get(j);
	    if (inBox( px, py, x+dc.m_x, y+dc.m_y, dc.m_width, dc.m_height )) {		
		m_clickHint = (dc instanceof CtrlConnector) ? "ctrl" : "data";
		m_clickedData = dc;
		return true;
	    }
	}

	// if click in contents of folder, set click hint here...
	return false;
    }

    boolean clickedTitle() { return m_clickHint.equals("title"); }
    boolean clickedResize() { return m_clickHint.equals("resize"); }
    boolean clickedResizeTL() { return m_clickHint.equals("resizetl"); }
    boolean clickedContents() { return m_clickHint.equals("contents"); }
    boolean clickedCtrl() { return m_clickHint.equals("ctrl"); }
    boolean clickedData() { return m_clickHint.equals("data"); }
    boolean clickedDataOrCtrl() { return clickedCtrl() || clickedData(); }


    // save children, return next oid
    int save(Writer out, String indent, int oid) throws IOException {
	out.write(indent+"<Folder label='"+m_label+"' x='"+x+"' y='"+y+"'>\n");
	
	// Things.
	Iterator<Thing> iter = m_things.iterator();
	while (iter.hasNext()) {
	    Thing t = iter.next();
	    oid ++;
	    if (t instanceof Place) {
		Place p = (Place)t;
		p.id = oid;
		String pstr = "  <Place id='"+oid+"' x='"+p.x+"' y='"+p.y+"'";
		if (p.isState) {
		    pstr = pstr + " state='true'";
		}
		if (p.m_outputCmd != null) {
		    pstr = pstr + " output='"+p.m_outputCmd+"'";
		}
		if (p.m_label != null) {
		    pstr = pstr + " label='"+p.m_label+"'";
		}
		
		out.write(pstr+"/>\n");
	    }
	    else {
		Folder f = (Folder)t;
		oid = f.save(out, indent+"  ", oid);
	    }	
	}
	
	// Connections
	iter = m_things.iterator();
	while (iter.hasNext()) {
	    Thing t = iter.next();
	    if (t instanceof Place) {
		Place p = (Place)t;
		if (p.links_out != null) {
		    for (int k=0; k<p.links_out.size(); k++) {
			Transition link = p.links_out.get(k);
			Place kid = link.m_to;		    
			out.write("  <Trans from='"+p.id+"' to='"+kid.id+"'");
			if (link.isOnset()) {
			    out.write(" onset='true'");
			}
			if (link.isDone()) {
			    out.write(" done='true'");
			}
		    
			out.write(">\n"); // transition
			// *** influences... ***
			out.write("  </Trans>\n");
		    }
		}
	    }
	}

	out.write("</Folder>\n");

	return oid;
    }


    Place getOrMakePlace(HashMap<Place,Place> idx, ArrayList<Thing> lst, Place oldPlace) {
	if (idx.containsKey(oldPlace)) { 
	    return idx.get(oldPlace); 
	}
	else {
	    Place np = oldPlace.copy();
	    idx.put(oldPlace, np);
	    lst.add(np);
	    return np;
	}
    }

    // interpolate places...
    void interp() {
	// make more places....
	ArrayList<Thing> iplace = new ArrayList<Thing>();

	// hash of old place to new place.
	HashMap<Place,Place> nplace = new HashMap<Place,Place>();

	Iterator<Thing> iter = m_things.iterator();
	while (iter.hasNext()) {
	    Thing t = iter.next();
	    if (t instanceof Place) {
		Place p = (Place)t;
		Place np = getOrMakePlace(nplace, iplace, p);		

		if (p.links_out != null) {
		    for (int k=0; k<p.links_out.size(); k++) {
			Transition out = p.links_out.get(k);
			Place kid = out.m_to;
			if (kid != null) { // null if link to ctrl connector....
			    int dist = kid.dist(p);
			    int xd = kid.x - p.x;
			    int yd = kid.y - p.y;
			    int diam = p.r*2;
			    int num = dist / diam;
			    Place lastPar = np;
			    
			    if (!out.isOnset() && !out.isDone()) {
				
				for (int j=1; j < num; j++) {
				    Place ip = new Place(p.x +j*xd/num, p.y+j*yd/num);
				    iplace.add( ip );
				    ip.m_parent = p.m_parent;
				    
				    Transition link = new Transition( lastPar, ip);
				    
				    ip.addIncoming(link);
				    lastPar.addOutgoing(link);
				    lastPar = ip;
				}
			    }
			    Place nkid = getOrMakePlace(nplace, iplace, kid);
			    
			    Transition link = new Transition(lastPar, nkid );
			    // copy onsetness
			    link.setOnset(out.isOnset());
			    link.setDone(out.isDone());
			    
			    // TBD *** do something abouut influences...
			    
			    nkid.addIncoming(link);
			    lastPar.addOutgoing(link);
			}
		    }
		}
	    }
	    else {
		iplace.add(t); // leave folders well alone for now...
	    }
	}      

	// *** Inter-folder links???
	
	// do children independently...
	iter = m_things.iterator();
	while (iter.hasNext()) {
	    Thing t = iter.next();
	    if (t instanceof Folder) {
		((Folder)t).interp();
	    }
	}

	m_things = iplace;
    }    
 
    // Reset animation to initial state (recursive)
    void resetAnim(boolean activate) {	
	if (m_behaviour != null) {
	    m_behaviour.reset();
	}

	Iterator<Thing> iter = m_things.iterator();
	while (iter.hasNext()) {
	    Thing t = iter.next();	    
	    if (t instanceof Place) {
		Place p = (Place)t;
		if (activate) {
		    p.initDistFromActive = 0; // activate 1st.
		    activate = false;
		}
		p.distFromActive = p.initDistFromActive;
		
		if (p.distFromActive == 0) {
		    p.updateDistFromActive(0);		   
		}
	    }
	    else if (t instanceof Folder) {
		((Folder)t).resetAnim(activate); // only activate top level thing...
	    }
	}	
    }    

    int countThings() {
	int ret = 0;
	Iterator<Thing> iter = m_things.iterator();
	while (iter.hasNext()) {
	    Thing t = iter.next();	    
	    ret ++;
	    if (t instanceof Folder) {
		ret += ((Folder) t).countThings();
	    }
	}
	return ret;	
    }

    // spread activity outwards, return alive count.
    int spreadActivity() {
	int alive = 0;
	Iterator<Thing> iter = m_things.iterator();
	while (iter.hasNext()) {
	    Thing t = iter.next();	    
	    if (t instanceof Place) {
		Place p = (Place)t;

		if (p.distFromActive == 0) {
		    p.updateDistFromActive(0);
		    alive ++;
		}
	    }
	    else if (t instanceof Folder) {
		alive += ((Folder) t).spreadActivity();
	    }
	}
	return alive;
    }

    void spreadFood() {
	Iterator<Thing> iter = m_things.iterator();
	while (iter.hasNext()) {
	    Thing t = iter.next();	    
	    if (t instanceof Place) {
		Place p = (Place)t;
	    
		if (p.distFromFood == 0) {
		    p.updateDistFromFood(0);
		}
	    }
	    else if (t instanceof Folder) {
		((Folder) t).spreadFood();
	    }
	}
    }


    boolean m_activityMoved = false;

    // sample the match on the previous cycle...
    boolean m_matched = false;
	
    void moveActivity() {
	if (!m_clk.enabled()) {
	    // sample clock on this phase...
	    System.out.println("Folder "+m_label+" not doing next state as no clock");
	    m_activityMoved = false;
	    return;
	}

	m_matched = m_match.enabled();
	if (m_matched) {
	    System.out.println("**** FOUND A MATCH IN "+m_label);
	}
	
	m_activityMoved = true;

	Iterator<Thing> iter = m_things.iterator();
	while (iter.hasNext()) {
	    Thing t = iter.next();	    
	    if (t instanceof Place) {
		Place p = (Place)t;
	    
		if (p.distFromActive == 0) {
		    p.moveActivity();
		}
	    }
	    else if (t instanceof Folder) {
		((Folder) t).moveActivity();
	    }
	}
    }

    // Copy next state to curr state
    void nextState() {  
	if (!m_activityMoved) return;

	Iterator<Thing> iter = m_things.iterator();
	while (iter.hasNext()) {
	    Thing t = iter.next();	    
	    //  System.out.println(" ... updating "+t);
	    
	    if (t instanceof Place) {
		Place p = (Place)t;

		// Do outputs on trigger...
		if (p.nextDistFromActive==0 && p.distFromActive != 0) {
		    if (p.m_outputCmd !=null) {
			if ("Text".equals(p.m_outputCmd)) {
			    System.out.print(p.m_label);
			    System.out.print(" ");
			}
			else if ("Note".equals(p.m_outputCmd)) {
			    int n = p.y;
			    if (p.m_parent != null) {				
				n = p.y - p.m_parent.y;
			    }
			    
			    if (p.m_label!=null) {
				n = Character.getNumericValue(p.m_label.charAt(0)) & 0xff;
				n = 50 -n;
			        n *= 6;
				System.out.println("note!!! "+p.m_label+" "+n);	
				if (p.m_label.equals(" ")) { 
				    n = 300;
				}
				if (p.m_label.length() > 1) { n=-1; }
			    }
			    
			    if (n!=-1) {
				Slime.startNote(n);
			    }
			    //			    System.out.println("note!!! " + n+" started");			    
			}
		    }
		}
	    
		p.distFromActive = p.nextDistFromActive;
		p.nextDistFromActive = -1; // reset for next time
	    }
	    else if (t instanceof Folder) {
		((Folder) t).nextState();
	    }

	}	

    }    


    // Check if reset signal applies
    void checkReset() {
	if (m_reset.enabled() && m_activityMoved) {
	    System.out.println(">>> Folder "+m_label+" performing a reset.");

	    // set activity to init...
	    Iterator<Thing> iter = m_things.iterator();
	    while (iter.hasNext()) {
		Thing t = iter.next(); 
		if (t instanceof Place) {
		    Place p = (Place)t;
		    // p.nextDistFromActive = p.initDistFromActive;
		    p.distFromActive = p.initDistFromActive;
		    // System.out.println(">>> Folder "+m_label+" setting "+p.initDistFromActive);
		}
	    }
	}
	else {
	    // System.out.println(">>> Reset not enabled for folder "+m_label);
	}
	  
	// moved the clocking here... after the nextState() called.
	if (m_activityMoved && m_behaviour != null) {
	    m_behaviour.clock();
	}


	Iterator<Thing> iter = m_things.iterator();
	while (iter.hasNext()) {
	    Thing t = iter.next(); 
	    if (t instanceof Folder) {
		((Folder) t).checkReset();
	    }
	}
    }


}


/**
 * A place.  A place has x,y coords.
 * and a "dist from active" setting - 0 == active, -1 == inactive, [1-8] == nearly active
 */
class Place extends Thing {
    static int r = 8;
    int id = -1;
    String m_outputCmd = null; // e.g. "Note"    

    boolean isState = false;

    // dist from activation (0 = active, 1 = on next cycle, -1 = not active)
    int distFromActive = -1; 
    int nextDistFromActive = -1;
    int initDistFromActive = -1; // initial dist from active

    // should we be building in a "food" marker here???
    // if so, why not also a "danger" flag?
    int distFromFood = -1;

    LinkedList<Transition> links_in = null; // past
    LinkedList<Transition> links_out = null; // future
    
    boolean m_needToDraw = true;

    Place() {
	this(0,0);
    }

    public int hashCode() { 
	return y * 1000 + x;
    }

    Place(int px, int py) {
	x = px;
	y = py;
    }        
    
    // shallow copy.
    Place copy() { 
	Place p = new Place(x,y); 
	p.m_label = m_label;
	p.initDistFromActive = initDistFromActive;    
	p.m_outputCmd = m_outputCmd;
	p.isState = isState;
	p.m_parent = m_parent;
	return p; 
    }

    void setIsState(boolean s) {
	isState = s;
    }
    
    void activateNext() {
	m_needToDraw = true;
	nextDistFromActive = 0;
	// signal any output onsets...
	if (links_out != null) {
	    Iterator<Transition> iter = links_out.iterator();
	    while (iter.hasNext()) {
		Transition trans = iter.next();
		if (trans.isOnset()) {
		    if (trans.m_to != null) {
			trans.m_to.activateNext();
		    }
		}
	    }
	}
    }
    
    // calc outgoing wave
    void updateDistFromActive(int depth) {
	if (depth > 5) return;

	nextDistFromActive = depth;
	m_needToDraw = true;

	// propogate decaying term recursively...
	if (links_out != null && !isState) {
	    for (int i=0; i<links_out.size(); i++) {
		Transition link = links_out.get(i);
		if (link.m_to != null) {
		    int ot = link.m_to.nextDistFromActive;
		    
		    if (ot==-1 || ot > depth) {
			if (!link.isOnset() && !link.isDone()) {
			    link.m_to.nextDistFromActive = depth;
			    link.m_to.updateDistFromActive(depth+1);
			}
		    }
		}
	    }
	}
    }

    // calc incoming dist from food
    void updateDistFromFood(int depth) {
	// NOP for now.
    }
  
    // move activity from one place to the next
    void moveActivity() {
	if (distFromActive == 0) {
	    // System.out.println("Moving place activity");
	    boolean propogate = true;
	    if (isState) {
		propogate = false;
		nextDistFromActive = 0; // stay on... don't trigger activate next....

		// decide whether to switch self off & propogate...
		if (links_in != null) {
		    Iterator<Transition> iter = links_in.iterator();
		    while (iter.hasNext()) {
			Transition trans = iter.next();
			if (trans.isDone() && trans.m_from.distFromActive==0) {
			    propogate = true;
			}
		    }
		}
	    }

	    if (propogate) {
		// move on....
		nextDistFromActive = -1;	
		m_needToDraw = true;
		moveOn();
	    }
	}		
    }

    static Random m_rand = new Random();

    void moveOn() {
	if (links_out != null) {
	    //	    double[] output_prob = new double[links_out.size()];
	    double[] output_cdf = new double[links_out.size()];
	    double tot = 0;
	    for (int i=0; i<links_out.size(); i++) {
		Transition trans = links_out.get(i);
		if (!trans.isDone() && !trans.isOnset()) {
		    double prob = trans.getProb();		  
		    tot += prob;
		    output_cdf[i] = tot;
		}
		else {
		    output_cdf[i] = -1;
		}		
	    }

	    // pick one at random... weighted by probabilities...
	    if (tot==0) return;
	    else {
		double val = m_rand.nextDouble() * tot;
		for (int i=0; i<output_cdf.length; i++) {
		    if (val <= output_cdf[i]) {
			// move to i !!!
			Place kp = links_out.get(i).m_to;
			kp.activateNext();
			return;
		    }
		}
	    }
	}
    }

    // returns boolean intersects w bbox
    boolean intersects(int px, int py) {
	int dx = x-px; if (dx<0) dx = -dx;
	int dy = y-py; if (dy<0) dy = -dy;

	if (dx > r+5 || dy > r+5) return false;
	return true;
    }

    // true if this place lies in the given box
    boolean inBox(int x1, int y1, int x2, int y2) {
	x1 -= r;
	x2 += r;
	y1 -= r;
	y2 += r;
	return x >= x1 && x <= x2 && y >= y1 && y <=y2;
    }

    // does it intersect main part or just bb?
    boolean intersectsCentre(int px, int py) {
	int dist = (int) Math.sqrt((px -x)*(px-x) + (py-y)*(py-y));
	if (dist < r) {
	    return true;
	}
	return false;
    }   

    int dist(Place p2) {
	int xd = x - p2.x;
	int yd = y - p2.y;
	return (int) Math.sqrt( xd*xd + yd*yd );
    }
    
    void addIncoming(Transition from) {
	if (links_in==null) {
	    links_in= new LinkedList<Transition>();
	}
	links_in.add(from);
    }

    void addOutgoing(Transition to) {
	if (links_out==null) {
	    links_out = new LinkedList<Transition>();
	}
	links_out.add(to); 
    }

    // incoming "onset" link... normal places, set t==0
    void notifyOnset() {
	distFromActive = 0;
    }

    void removeAllFrom(LinkedList<Transition> list, Place old) {
	if (list != null) {
	    Iterator<Transition> iter = list.iterator();
	    while (iter.hasNext()) {
		Transition t = iter.next();
		if (t.m_from == old || t.m_to==old) {
		    iter.remove();
		}
	    }
	}
    }

    // another place has been deleted - delete references to it
    // *** This requires a complete iteration through the lists ---
    // would be faster to use backwards links...
    void deleteReferencesTo(Place old) {
	removeAllFrom(links_in,old);
	removeAllFrom(links_out,old);
    }


    static Color m_normalCol = new Color(200,210,200,255);
    static Color m_selCol    = new Color(200,0,0,150);

    void paint(Graphics gg, MapPanel p, boolean force) {
			
	if (!force && !m_needToDraw) {
	    return;
	}

	// pick a colour depending on t...
	int dt = distFromActive; // - m_currT;
	    
	if (isState) {
	    if (distFromActive==0) {
		gg.setColor(Color.magenta);
	    }
	    else {
		gg.setColor(Color.magenta.darker());
	    }
	    gg.fillRect(p.getX(x-r), p.getY(y-r), p.getE(x-r,r*2+1), p.getE(y-r,r*2+1));
	}
	
	if (dt >= 0 && dt < 10) {
	    gg.setColor(p.getCol(dt) );
	    gg.fillOval(p.getX(x-r), p.getY(y-r), p.getE(x-r,r*2), p.getE(y-r,r*2));
	}	    
	else {
	    if (m_needToDraw) {
		gg.setColor(MapPanel.m_bg);
		gg.fillOval(p.getX(x-r), p.getY(y-r), p.getE(x-r,r*2), p.getE(y-r,r*2));
	    }

	    if (m_sel) {
		gg.setColor(m_selCol);
	    }
	    else {
		gg.setColor(m_normalCol);
	    }
	    gg.drawOval(p.getX(x-r), p.getY(y-r), p.getE(x-r,r*2), p.getE(y-r,r*2));
	}
	
	/*
	if (m_sel) {
	    gg.setColor(Color.red);
	    gg.fillOval(p.getX(x-r), p.getY(y-r), p.getE(x-r,r*2), p.getE(y-r,r*2));		
	    }*/
	
	// do a green blob in initial place
	if (initDistFromActive == 0) {
	    gg.setColor(Color.green);
	    gg.fillOval(p.getX(x-r/2), p.getY(y-r/2), p.getW(r), p.getW(r));		
	}
	
	if (m_outputCmd != null) {
	    if ("Note".equals(m_outputCmd)) {
		gg.setColor(Color.blue); 
	    }
	    else {
		gg.setColor(Color.white); 
	    }
	    gg.fillRect(p.getX(x), p.getY(y), p.getW(r), p.getW(r));
	}
	
	if (m_label != null) {
	    gg.setColor(Color.yellow);
	    gg.drawString(m_label, p.getX(x-r - r), p.getY(y-r));
	}
	
	m_needToDraw = false;
		
	// draw to children...
	if (links_out != null) {
	    Iterator<Transition> iter = links_out.iterator();
	    while (iter.hasNext()) {
		Transition trans = iter.next();
		Place kid = trans.m_to;
		CtrlConnector kidc = trans.m_toCtrl;

		int kx,ky;
		if (kid != null) {
		    kx = kid.x;
		    ky = kid.y;
		}
		else {
		    kx = kidc.getCentreX() + kidc.getParentX();
		    ky = kidc.getCentreY() + kidc.getParentY();
		}

		
		int xdist = kx - x;
		int ydist = ky - y;
		

		// draw a blob for the transition?
		int mind = r*4;
		int blobx = (x + kx) / 2;
		int bloby = (y + ky) / 2;
		
		if (xdist > mind || xdist < -mind || ydist > mind || ydist < -mind) {
		    // Yes! draw a blob...
		    int blobw = 5;
		    gg.setColor(Color.blue);
		    gg.fillRect(p.getX(blobx-blobw/2), p.getY(bloby-blobw/2), p.getW(blobw), p.getW(blobw));			
		    
		    // ... make it sensitive?
		    // m_blobs.add(...)
		    trans.setPos( blobx, bloby, blobw );
		    if (m_parent != null) m_parent.m_visibleTransitions.add(trans);
		}
		
		if (trans.isOnset() || trans.isDone()) {
		    if (trans.isDone()) { 
			gg.setColor(Color.red.darker()); 
		    }
		    else { 
			gg.setColor(Color.cyan.darker());
		    }
		    gg.drawLine(p.getX(kx),  p.getY(ky), p.getX(x), p.getY(y) ); 
		}
		else {
		    gg.setColor(Color.black);
		    gg.drawLine(p.getX( kx),  p.getY(ky), p.getX(x), p.getY(y) ); 
		    
		    gg.setColor(Color.magenta);
		    if (kid != null) {
			int dist = kid.dist(this);
			if (dist<3) dist = 3;
			int subdist = dist -3;
			gg.drawLine( p.getX(x + ((kx-x) * subdist / dist)),  p.getY(y + ((ky-y) * subdist / dist)), 
				     p.getX(kx), p.getY(ky));
		    }
		}
		
		// draw influences???
		if (trans.isInfluenced()) {
		    gg.setColor(INFL_COL);
		    // and dotted?
		    Iterator<Influence> inf = trans.m_influencedBy.iterator();
		    while (inf.hasNext()) {
			Influence infi = inf.next();
			gg.drawLine(p.getX(blobx),  p.getY(bloby), p.getX(infi.getSrcX()), p.getY(infi.getSrcY()) ); 
		    }
		}
	    }	       
	}
	
    }

    static Color INFL_COL = new Color(80,80,80);
}

interface Behaviour {
    void clock();
    void reset();
}


// a pattern recorder
class Recorder implements Behaviour {
    Folder m_folder;
    int m_count;
    int m_xstep = 40;

    int m_timeY = 15;

    Place m_prevPlace = null;

    Recorder(Folder folder) {
	m_folder = folder;
	reset();
    }

    public void clock() {
	if (m_count > 50) return; // capacity limit...
	
	if (m_folder.m_reset.enabled()) {
	    System.out.println("*** Recorder reset .. checking match");
	    if (!m_folder.m_matched) {
		System.out.println("*** Recorder reset, with no match ticked");
		
		SwingUtilities.invokeLater(new Runnable() { public void run() {
		    System.out.println("Actually duplicating things..");
		    // OK - remove our behaviour.
		    m_folder.setBehaviour(null); // or a recognizer?
		    m_folder.m_label="-"; // fixed...

		    // Make a new "Match" place
		    m_count ++;
		    int xp = m_count*m_xstep+m_folder.x + 20;
		    Place match = new Place(xp, m_timeY+m_folder.y);
		    m_folder.addThing(match);
		    if (m_prevPlace != null) {
			Transition ct = new Transition(m_prevPlace, match);
			m_prevPlace.addOutgoing(ct);
			match.addIncoming(ct);
			ct.addInfluence( new Influence( m_folder.m_tm, false, false ) );
		    }
		    
		    // Join the match place to the m_match output.
		    Transition mo = new Transition(match, m_folder.m_match);
		    mo.setOnset(true);
		    match.addOutgoing(mo);
		    m_folder.m_match.addIncoming(mo);

		    Junction matchJunction = null;

		    // Was there no incoming match junction?
		    if (m_folder.m_match.m_link != null) {
			DataConnector fc = m_folder.m_match.m_link.m_from;
			if (fc.m_parent instanceof Junction) {
			    matchJunction = (Junction)fc.m_parent;
			    matchJunction.removeOutputs();			    
			    // remove old link...
			    m_folder.m_parent.removeThing(m_folder.m_match.m_link);
			    m_folder.m_match.m_link = null;
			}			
		    }

		    // If no match junction, make one.
		    if (matchJunction == null) {			
			matchJunction = new Junction();	
			matchJunction.setCtrl();
			m_folder.m_parent.addThing(matchJunction);
			matchJunction.x = m_folder.x + m_folder.m_match.getCentreX() + 60;
			matchJunction.y = m_folder.y + m_folder.m_match.getCentreY();
		    }
		    
		    // add a link from match output to junction.
		    DataConnector jin = matchJunction.makeInput();
		    DataLink mlink = new DataLink(m_folder.m_match, jin);
		    mlink.setCtrl();
		    m_folder.m_parent.addThing(mlink);

		    // Make a new folder - a duplicate of m_folder.
		    int dy = m_folder.m_height+55;
		    int dx = 0;
		    Folder newFolder = new Folder("recorder", m_folder.x+dx, m_folder.y + dy);
		    newFolder.m_width = m_folder.m_width;
		    newFolder.m_height = m_folder.m_height;
		    newFolder.layoutConnectors();
		    newFolder.setBehaviour(new Recorder(newFolder));	    
		    m_folder.m_parent.addThing(newFolder);
		    
		    newFolder.copyConnectionsFrom(m_folder, dx, dy);

		    // Add incoming link for "match" input.
		    Junction matchInJunction = new Junction();
		    matchInJunction.setCtrl();
		    matchInJunction.x = matchJunction.x;
		    matchInJunction.y = newFolder.y + newFolder.m_match.getCentreY();
		    m_folder.m_parent.addThing(matchInJunction);
		    DataConnector j1out = matchJunction.makeOutput();	
		    j1out.m_xedge = 0; j1out.m_yedge=1;
		    DataConnector j2in = matchInJunction.makeInput();
		    j2in.m_xedge = 0; j2in.m_yedge=-1;
		    DataLink dlink1 = new DataLink(j1out, j2in);
		    dlink1.setCtrl();
		    m_folder.m_parent.addThing(dlink1);
		    
		    // Make leftwards connection from new junction to new match input.
		    DataConnector j2out = matchInJunction.makeOutput();
		    j2out.m_xedge = -1; j2out.m_yedge = 0;
		    DataLink dlink2 = new DataLink(j2out, newFolder.m_match);
		    dlink2.setCtrl();
		    m_folder.m_parent.addThing(dlink2);
		    
		    		    		    
		    
		}});
	    }
	    else {
		System.out.println("*** Behaviour folder performing a reset because match found");
		reset();
	    }

	    return;
	}

	m_count++;

	// maybe do something...

	int xp = m_count*m_xstep+m_folder.x + 20;

	// Make a new time place anyway...
	Place tp = new Place(xp, m_timeY+m_folder.y);
	if (m_count==1) {
	    tp.initDistFromActive = 0;
	}
	m_folder.addThing(tp);
	if (m_prevPlace != null) {
	    Transition ct = new Transition(m_prevPlace, tp);
	    m_prevPlace.addOutgoing(ct);
	    tp.addIncoming(ct);
	    m_prevPlace = tp;

	    // add the dependency ...
	    ct.addInfluence( new Influence( m_folder.m_tm, false, false ) );
	}
	else {
	    m_prevPlace = tp;
	}

	LinkedList<Folder> sources = m_folder.getInputFolders();
	for (Folder src : sources) {
	    // check inputs of folder in case there's anything good to add.
	    // could check whether folder is output enabled...
	    Iterator<Thing> iter = src.m_things.iterator();
	    if (iter != null) {
		while (iter.hasNext()) {		
		    Thing t = iter.next();
		    if (t instanceof Place) {
			Place p = (Place)t;
			if (p.distFromActive == 0 && p.m_outputCmd != null) {
			    // System.out.println("rec clocked by adding place");			
			    // we have activity!  Make a new place!!	
			    
			    int ypos = m_folder.y + p.y;
			    if (p.m_parent!=null) { ypos -= p.m_parent.y; }	
			    if (p.m_label!=null) {
				ypos = 60 + m_folder.y; // if label set, fix the output
			    }
			    
			    Place newPlace = new Place(xp, ypos);
			    newPlace.m_outputCmd = "Note";
			    newPlace.m_label = p.m_label;

			    if (m_count==1) {
				newPlace.initDistFromActive = 0;
			    }
			    
			    m_folder.addThing(newPlace);
			    // add a transition from time place...
			    Transition onset = new Transition(tp, newPlace);
			    onset.setOnset(true);
			    tp.addOutgoing(onset);
			    newPlace.addIncoming(onset);
			}		   
		    }
		}
	    }
	    else {
		System.out.println("rec clocked by couldn't find input");
	    }
	}
    }

    public void reset() {
	m_count = 0;
	m_prevPlace = null;
	m_folder.m_things = new ArrayList<Thing>();
    }
}


class KeyboardReader implements Behaviour {
    Folder m_folder;
    MapPanel m_panel;

    public KeyboardReader(Folder f, MapPanel panel) {
	m_folder = f;
	m_panel = panel;
	reset();
    }

    Place findPlace(String lab) {
	for (int i=0; i<m_folder.m_things.size(); i++) {
	    Thing t =(Thing) m_folder.m_things.get(i);
	    if (t instanceof Place) {
		if (lab.equals(((Place)t).m_label)) {
		    return (Place)t;
		}
	    }
	}
	return null;
    }
 
    public void pressed(char c) {
	System.out.println("char "+c+" pressed");
	Place p = findPlace(""+Character.toUpperCase(c));
	if (p!=null) {
	    p.m_needToDraw = true;

	    p.nextDistFromActive = 0; // maybe deactivate all others???
	    Place op = findPlace("onset");
	    if (op != null) {
		// set the DIST FROM ACTIVE so the onset can be used as a clock.
		findPlace("onset").distFromActive = 0;
		findPlace("onset").nextDistFromActive = -1;
	    }
	    m_panel.move();
	    m_panel.m_force = true;
	    m_panel.repaint(); // paint();
	}
    }

    public void released(char c) {
	System.out.println("char "+c+" released");
	Place p = findPlace(""+Character.toUpperCase(c));
	if (p!=null) {
	    p.m_needToDraw = true;
	    //p.nextDistFromActive = -1; // maybe deactivate all others???
	    //findPlace("onset").nextDistFromActive = -1;
	    m_panel.move();
	    m_panel.m_force = true;
	    m_panel.repaint();
	}
    }

    public void clock() {
    }

    void addKeys(String str, int x, int y) {
	for (int i=0; i<str.length(); i++) {
	    Place kp = new Place(x + i * 32 + m_folder.x, y + m_folder.y);
	    kp.m_label=""+str.charAt(i);
	    kp.m_outputCmd="Note";
	    m_folder.addThing(kp);
	}
	m_panel.repaint();
    }

    public void reset() {
	if (m_folder.m_things.size()<26) {
	    m_folder.m_things = new ArrayList<Thing>();
	    System.out.println("Keyboard reset called");
	    // add the a-z keys...
	    addKeys("QWERTYUIOP", 32, 60);
	    addKeys("ASDFGHJKL", 48, 100);
	    addKeys("ZXCVBNM.?", 64, 140);
	    addKeys(" ", 200, 180);
	    Place onset = new Place(400+m_folder.x, 200+m_folder.y);
	    onset.m_label="onset";
	    m_folder.addThing(onset);	
	    m_folder.expandToFit(onset);
	    m_folder.layoutConnectors();
	}	 
    }
}

class KeyboardAdapter extends KeyAdapter {
    Slime m_parent;

    KeyboardAdapter(Slime slime) {
	m_parent = slime;
    }

    public void keyPressed(KeyEvent e) {
	if (m_parent.m_keyboardReader != null) {
	    m_parent.m_keyboardReader.pressed(e.getKeyChar());
	}
    }
    public void keyReleased(KeyEvent e) {
	if (m_parent.m_keyboardReader != null) {
	    m_parent.m_keyboardReader.released(e.getKeyChar());
	}
    }    
}


// panel for a map.
class MapPanel extends JPanel {

    double m_zoom = 1.0;

    int m_width = 8000; // size, in base coords
    int m_height = 2000;

    int m_but;

    Folder m_root = new Folder("root", 0,0);
    Folder m_context = m_root;

    Anim m_anim;

    Slime m_slime;

    Thing m_thing1, m_thing2;
    Transition m_trans1, m_trans2;
    DataLink m_dataLink;

    boolean m_first = true;
    double ZOOM = 1.2;

    int m_offx, m_offy; // offset into thing when dragging
    int m_boxx, m_boxy;
    int m_boxx2, m_boxy2;
    boolean m_drawBox = false;

    boolean m_force = true; // whether to force a repaint...
    boolean m_reallyForce = false; 

    // zoom in keeping x,y in same place on screen
    // x,y in global coords.
    void zoomIn(int x, int y) {
	if (m_zoom > 0.1) {
	    
	    JScrollPane scr = m_slime.m_scroll;
	    Dimension vp = scr.getViewport().getExtentSize();
	    Point tl = scr.getViewport().getViewPosition();

	    int x1 = pixToX(tl.x); // initial position of top left
	    int y1 = pixToY(tl.y);


	    int x2 = x - (int) ((x-x1)*ZOOM);
	    int y2 = y - (int) ((y-y1)*ZOOM);

	    System.out.println("zooming in at "+x+","+y+"   x1y1 "+x1+","+y1+"   x2y2 "+x2+","+y2);
	    
	    m_zoom /= ZOOM;
	    setPreferredSize( new Dimension( (int) (m_width * m_zoom), (int) (m_height * m_zoom) ));
	    int xx = getX(x2); if (xx<0) xx = 0;
	    int yy = getY(y2); if (yy<0) yy = 0;
	    scr.getViewport().setViewPosition(new Point(xx, yy));
	    m_force = true;
	    revalidate();
	    // repaint();
	}
    }

    void zoomOut(int x, int y) {
	if (m_zoom < 10) {
	    JScrollPane scr = m_slime.m_scroll;
	    Dimension vp = scr.getViewport().getExtentSize();
	    Point tl = scr.getViewport().getViewPosition();

	    int x1 = pixToX(tl.x); // initial position of top left
	    int y1 = pixToY(tl.y);

	    int x2 = x - (int) ((x-x1)/ZOOM);
	    int y2 = y - (int) ((y-y1)/ZOOM);

	    if (x2<0) x2 = 0;
	    if (y2<0) y2 = 0;

	    // System.out.println("zooming out at "+x+","+y+"   x1y1 "+x1+","+y1+"   x2y2 "+x2+","+y2);

	    m_zoom *= ZOOM;
	    setPreferredSize( new Dimension( (int) (m_width * m_zoom), (int) (m_height * m_zoom) ));

	    int xx = getX(x2); if (xx<0) xx = 0;
	    int yy = getY(y2); if (yy<0) yy = 0;
	    scr.getViewport().setViewPosition(new Point(xx, yy));

	    //	    scr.getViewport().setViewPosition(new Point(getX(x2), getY(y2)));
	    // scr.getViewport().setViewPosition(new Point((int) (x2*m_zoom), (int)(y2*m_zoom)));
	    m_force = true;
	    revalidate();
	    //repaint();
	}
    }

    LinkedList<Thing> getSelectedThings() {
	LinkedList<Thing> ret = new LinkedList<Thing>();
	m_root.addToSelectedThings(ret);
	return ret;
    }

    void selectOnly(Thing thing) {
	m_slime.updatePlacePanel(thing);
	m_root.selectOnly(thing);
    }

    void selectBox(int x1, int y1, int x2, int y2) {
	m_root.selectBox(x1, y1, x2, y2);
    }

    void deleteSelected() {
	LinkedList<Thing> deleted = new LinkedList<Thing>();
	m_root.deleteSelected(deleted, false);
	m_root.deleteReferencesTo(deleted);
	repaint();
    }

    int folderNum = 0;

    // Add a new folder...
    void groupSelected() {	
	Folder f = m_root.groupSelected(null);
	repaint();
    }    

    void ungroupSelected() {
	m_root.ungroupSelected();
	repaint();
    }    

    // find folder parent at given point.
    Folder getContextAt(int px, int py) {
	return m_root.findContext(px, py);
    }

    // Make the main panel.
    MapPanel(Slime parent) {
	super();
	
	m_slime = parent;

	addMouseListener(new MouseInputAdapter() {		
		boolean m_zooming = false;

		boolean m_fromBorder = false; // link from state border

		// select or create a new place.
		public void mousePressed(MouseEvent me) {

		    // request focus.
		    requestFocusInWindow();

		    m_but = me.getButton();
		    int x = pixToX(me.getX());
		    int y = pixToY(me.getY());
		    m_dataLink = null; // reset this.
		    m_thing1 = m_root.findThing(x,y);
		    m_context = getContextAt(x,y);

		    System.out.println("click : found thing "+m_thing1);
		    m_trans1 = null;

		    if (m_thing1==null)
			m_trans1 = m_context.findTransition(x,y);	
	    
		    if (m_thing1==null && m_trans1==null) {
			if (m_but==1) {
			    m_thing1 = new Place(x,y);
			    m_context.addThing(m_thing1);			    
			}
			else if (m_but==3) {
			    m_thing1 = new Place(x,y);
			    ((Place)m_thing1).setIsState(true);
			    m_context.addThing(m_thing1);
			}
			else {
			    m_boxx = x;
			    m_boxy = y;
			    m_boxx2 = x + 1;
			    m_boxy2 = y + 1;
			    m_drawBox = true;			    
			}
		    }
		    if (m_thing1!=null) {
			selectOnly( m_thing1 );
			m_offx = x - m_thing1.getX();
			m_offy = y - m_thing1.getY();
		    }
		    
		    m_first = true;
		    repaint();
		}

		public void mouseReleased(MouseEvent me) {		    
		    // glue start place to end place...
		    int x = pixToX(me.getX());
		    int y = pixToY(me.getY());

		    if (m_but==1 || m_but==3) {
			if (m_thing1!=null && m_thing1.intersects(x,y)) 
			    return;

			m_thing2 = m_root.findThing(x,y);
			m_trans2 = null;

			Folder target = getContextAt(x,y);

			if (m_thing2 == null) {
			    m_trans2 = m_root.findTransition(x,y);
			}

			if (m_thing2 == null && m_trans2==null) {
			    if (m_dataLink != null) {
				// make a new junction
				Junction j = new Junction();
				if (m_dataLink.isCtrl()) {
				    j.setCtrl();
				}
				j.x = x;
				j.y = y;
				if (m_dataLink.m_from.m_xedge!=0) {
				    j.y = m_dataLink.m_from.getConnY() + m_thing1.y;
				}
				if (m_dataLink.m_from.m_yedge!=0) {
				    j.x = m_dataLink.m_from.getConnX() + m_thing1.x;
				}
				j.addInput(m_dataLink.m_to);
				// j.m_in = m_dataLink.m_to;

				// change parent...
				m_dataLink.m_to.changeParent( j );
				m_dataLink.setDragging(false);

				// fix sign of edge of in..
				m_dataLink.m_to.m_xedge = -m_dataLink.m_from.m_xedge;
				m_dataLink.m_to.m_yedge = -m_dataLink.m_from.m_yedge;
				
				m_thing2 = j;
				target.addThing(m_thing2);
				selectOnly(m_thing2);
				m_dataLink = null;
			    }
			    else {				
				// make a new place.
				m_thing2 = new Place(x,y);
				target.addThing(m_thing2);
				selectOnly(m_thing2);
			    }
			}	
			else if (m_thing2 instanceof Folder && m_dataLink != null) {
			    // dropped onto a folder...
			    Folder f = (Folder) m_thing2;

			    if (f.clickedData()) {
				// fussy, want to release on a input port.
				DataConnector dc = f.m_clickedData;
				m_dataLink.m_to = dc;
				dc.m_link = m_dataLink;
				// Jiggle position of target???
				//				System.out.println("Connected to "+dc);
				
				selectOnly(m_thing2);
				m_dataLink.setDragging(false);
				m_dataLink = null;
			    }
			    else if (f.clickedCtrl()) {
				// fussy, want to release on a input port. // check 
				CtrlConnector dc = (CtrlConnector)f.m_clickedData;
				m_dataLink.m_to = dc;
				dc.m_link = m_dataLink;
						
				selectOnly(m_thing2);
				m_dataLink.setDragging(false);
				m_dataLink = null;
			    }
			}
			else if (m_thing2 instanceof Junction && m_dataLink != null) {
			    // landed on junction.
			    Junction tgt = (Junction) m_thing2;
			    if (tgt.isCtrl() == m_dataLink.isCtrl()) {
				DataConnector inp = tgt.makeInput();
				m_dataLink.m_to = inp;
				m_dataLink.m_to.changeParent(tgt);
				inp.m_link = m_dataLink;
				selectOnly(m_thing2);
				m_dataLink.setDragging(false);
				m_dataLink = null;
			    }
			}
			
			if (m_thing1 instanceof Place && m_thing2 instanceof Place) {
			    Place place1 = (Place)m_thing1;
			    Place place2 = (Place)m_thing2;
			    
			    if (m_but==1) {
				Transition link = new Transition(place1, place2);
				place1.addOutgoing(link);
				place2.addIncoming(link);
			    }
			    else if (m_but==3) {
				Transition link = new Transition(place1, place2);
				if (place2.isState) {
				    link.setDone(true);
				}			
				else {
				    link.setOnset(true);
				}
				
				place1.addOutgoing(link);
				place2.addIncoming(link);
			    }
			}
			else if (m_thing1 instanceof Place && m_thing2 instanceof Folder) {
			    // place to ctrl connector??
			    Folder f2 = (Folder)m_thing2;

			    System.out.println("Linking a place to a control connector");
			    if (f2.clickedCtrl()) {
				Transition link = new Transition((Place)m_thing1, (CtrlConnector)f2.getClickedDataConnector());
				link.setOnset(true);
				((Place)m_thing1).addOutgoing(link);
				link.m_toCtrl.addIncoming(link);
			    }
			    
			}
			else if (m_trans1!=null) {
			    if (m_thing2 instanceof Place) {
				// make place influence transition
				m_trans1.addInfluence(new Influence((Place)m_thing2, false, false));
			    }
			    else if ((m_thing2 instanceof Folder) && (((Folder)m_thing2).clickedCtrl())) {
				Folder f2 = (Folder)m_thing2;
				m_trans1.addInfluence(new Influence((CtrlConnector)f2.getClickedDataConnector(), false, false));
			    }
			}
			else if (m_trans2!=null && m_thing1 instanceof Place) {
			    // make place influence transition
			    m_trans2.addInfluence(new Influence((Place)m_thing1, false, false));
			}
			

		    }
		    else if (m_but == 2) {
			if (m_thing1 == null) {
			    selectBox( m_boxx, m_boxy, x, y );
			    m_drawBox = false;
			}
		    }

		    repaint();
		}
	    });
	

	addMouseWheelListener(new MouseWheelListener() {
		public void mouseWheelMoved(MouseWheelEvent e) {
		    if (e.getScrollType() == MouseWheelEvent.WHEEL_UNIT_SCROLL) {
			int a = e.getUnitsToScroll();
			int x = pixToX(e.getX());
			int y = pixToY(e.getY());

			System.out.println("event pos "+e.getX()+" "+e.getY());			
			if (a<0) { 
			    zoomIn(x,y); 
			}
			else {
			    zoomOut(x,y);
			}
			System.out.println("Scrolling "+a+" units... zoom now "+m_zoom);
		    }
		}
	    });

	addMouseMotionListener(new MouseMotionAdapter() {
		int x1,y1,x2,y2;
		int ox = -1;
		int oy = -1;

		public void mouseMoved(MouseEvent me) {
		    m_reallyForce = true;
		    
		}

		public void mouseDragged(MouseEvent me) {
		    Graphics g = getGraphics();


		    Graphics2D g2d = (Graphics2D)g;
		    //g2d.setRenderingHint( RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON );

		    int xp = me.getX();		    
		    int yp = me.getY();

		    boolean draggingFolder = false;

		    if (m_thing1 instanceof Folder) {
			Folder f = (Folder)m_thing1;
			if (f.clickedTitle() || f.clickedResize() || f.clickedResizeTL()) {
			    draggingFolder = true;
			}
		    }

		    if (!draggingFolder &&( m_but==1 || m_but==3)) {			
			g.setXORMode(Color.gray);

			int thingx =0, thingy=0;
			if (m_thing1!=null) { thingx = m_thing1.x; thingy = m_thing1.y; }
			else if (m_trans1!=null) { thingx = m_trans1.x; thingy = m_trans1.y; }
			
			// Is it a fatline?
			if (m_thing1 instanceof Junction || (m_thing1 instanceof Folder && ((Folder)m_thing1).clickedDataOrCtrl())) {
				if (m_first) {
				    DataConnector from = m_thing1.getClickedDataConnector();
				    DataConnector to = from.makeComplement();
				    m_dataLink = new DataLink(from, to);
				    if (from instanceof CtrlConnector) {
					m_dataLink.setCtrl();
				    }

				    m_dataLink.setDragging(true);
				    m_thing1.m_parent.addThing(m_dataLink);
				    repaint();
				}
				
				// draw a fat line...
				else {
				    // move the edge if needed...
				    if (m_thing1 instanceof Junction) {
					int px = pixToX(xp);
					int py = pixToY(yp);
					// choose the direction....
					int dx = px - (m_dataLink.m_from.m_x + m_thing1.x);
					int dy = py - (m_dataLink.m_from.m_y + m_thing1.y);
					int adx = dx; if (adx<0) adx=-dx;
					int ady = dy; if (ady<0) ady=-dy;
					if (dx > 0 && adx >= ady) {
					    // right
					    m_dataLink.m_from.m_xedge = 1;
					    m_dataLink.m_from.m_yedge = 0;
					}
					else if (dx < 0 && adx >= ady) {
					    // left
					    m_dataLink.m_from.m_xedge = -1;
					    m_dataLink.m_from.m_yedge = 0;
					}
					else if (dy < 0 && adx <= ady) {
					    // U
					    m_dataLink.m_from.m_xedge = 0;
					    m_dataLink.m_from.m_yedge = -1;
					}
					else if (dy > 0 && adx <= ady) {
					    // down
					    m_dataLink.m_from.m_xedge = 0;
					    m_dataLink.m_from.m_yedge = 1;
					}
				    }
				    
				    if (m_dataLink.m_from.m_xedge != 0) {
					// left / right...
					m_dataLink.m_to.m_x = pixToX(xp) - m_thing1.x;
				    }
				    if (m_dataLink.m_from.m_yedge != 0) {
					m_dataLink.m_to.m_y = pixToY(yp) - m_thing1.y;
				    }
				    repaint();
				}
			}
			else {
			    if (!m_first) {
				g.drawLine(getX(thingx), getY(thingy), ox, oy);
			    }
			    g.drawLine(getX(thingx), getY(thingy), xp, yp);
			}
			ox = xp; oy = yp;
			m_first = false;
		    }
		    else if (draggingFolder || m_but==2) {
			if (m_thing1==null) {
			    // dragging a box...
			    g.setXORMode(Color.gray);

			    //if (!m_first) {
			    //g.drawRect(bx, by, ox-bx, oy-by);
			    //}
			    m_boxx2 = pixToX(xp);
			    m_boxy2 = pixToY(yp);
			    selectBox( m_boxx, m_boxy, m_boxx2, m_boxy2);
			    repaint();
			    //			    g.drawRect(bx, by, xp-bx, yp-by); 
			    
			    ox = xp; oy = yp;
			    m_first = false;
			}
			else {
			    m_root.clearShifted();

			    if (m_thing1 instanceof Folder) {
				Folder f = (Folder)m_thing1;
				if (f.clickedResize()) {
				    f.preResize(); // compute abs position of connectors
				    f.m_width  = pixToX(xp) - m_thing1.x;
				    f.m_height = pixToY(yp) - m_thing1.y;
				    f.m_shifted = true;
				    f.layoutConnectors();
				}
				else if (f.clickedResizeTL()) {
				    f.preResize(); // compute abs position of connectors
				    System.out.println("Resize top left of folder...");
				    f.m_width  -= pixToX(xp) - (m_thing1.x+m_offx);
				    f.m_height -= pixToY(yp) - (m_thing1.y+m_offy);

				    m_thing1.x = pixToX(xp) - m_offx;
				    m_thing1.y = pixToY(yp) - m_offy; // don't move contents...
				    f.m_shifted = true;
				    f.layoutConnectors();
				}
				else {
				    m_thing1.move(pixToX(xp) - m_offx, pixToY(yp) - m_offy);
				}
			    }
			    else {
				m_thing1.move(pixToX(xp) - m_offx, pixToY(yp) - m_offy);
			    }

			    m_root.clearShifted();

			    repaint();
			}
		    }
		}
	    });

	setPreferredSize(new Dimension(520,520));
    }

    void reset() {
	m_root = new Folder();
	repaint();
    }

    void save() {
	try {
	    File f = new File("out.xml");

	    BufferedWriter out = new BufferedWriter(new FileWriter(f));
	    m_root.save(out, "", 0);
	    
	    out.close();
	} catch (Exception ex) {
	    System.out.println("probs "+ex);
	    ex.printStackTrace();
	}
    }
    
    int atoi(String a) {
	try {
	    return Integer.parseInt(a);
	} catch (Exception ex) {
	    return -1;
	}
    }
    
    String getStringParam(String name) {	
	return xpp.getAttributeValue(null, name);
    }
    int getIntParam(String name) {
	return atoi(xpp.getAttributeValue(null, name));
    }

    boolean getBoolParam(String name) {
	return "true".equals(xpp.getAttributeValue(null, name));
    }

    XmlPullParser xpp;



    void load() {	
	try {
	File f = new File("out.xml");
	FileReader fr = new FileReader(f);
	reset();

	XmlPullParserFactory factory = XmlPullParserFactory.newInstance();
	factory.setNamespaceAware(false);

	m_root = null;
	Folder currf = m_root; // current folder.
	xpp = factory.newPullParser();
	xpp.setInput( fr );
	int eventType = xpp.getEventType();

	ArrayList<Thing> allthings = new ArrayList<Thing>(); // for quick indexing.

	while (eventType != XmlPullParser.END_DOCUMENT) {
	    if(eventType == XmlPullParser.START_DOCUMENT) {
		System.out.println("Start document");
	    } else if(eventType == XmlPullParser.END_DOCUMENT) {
		System.out.println("End document");
	    } else if(eventType == XmlPullParser.START_TAG) {
		String tag = xpp.getName();
		if ("Place".equals(tag)) {
		    Place p = new Place();
		    p.x = getIntParam("x");
		    p.y =  getIntParam("y");
		    p.id = getIntParam("id");
		    p.isState = getBoolParam("state");
		    p.m_outputCmd = getStringParam("output");
		    p.m_label = getStringParam("label");
		    currf.addThing(p);
		    allthings.add(p); // hope ids match with index...
		    if (allthings.size()-1 != p.id) {
			System.out.println("error - index is "+(allthings.size()-1)+" but id is "+p.id);
		    }
		}
		else if ("Folder".equals(tag)) {
		    Folder newf = new Folder();
		    if (currf == null) { 
			m_root = newf; 
		    }
		    else {
			currf.addThing(newf);
		    }
		    currf = newf; 
		    currf.x = getIntParam("x");
		    currf.y = getIntParam("y");
		    currf.m_label = getStringParam("label");
		    allthings.add(currf); // hope ids match with index...
		}
		else if ("Trans".equals(tag)) {
		    int from = getIntParam("from");
		    int to = getIntParam("to");
		    // assume ids line up ok...
		    Thing ft = allthings.get(from);
		    Thing tt = allthings.get(to);

		    if (ft instanceof Place && tt instanceof Place) {
			Place fp = (Place)ft;
			Place tp = (Place)tt;
			
			Transition link = new Transition(fp, tp);
			if (getBoolParam("onset")) {
			    link.setOnset(true);
			}
			if (getBoolParam("done")) {
			    link.setDone(true);
			}
			
			fp.addOutgoing(link);
			tp.addIncoming(link);
		    }
		    else {
			System.out.println("Problem with transition - ids not places");
		    }
		    // plus influences...
		}
		else {
		    System.out.println("Start tag "+tag+" not known");
		}
	    } else if(eventType == XmlPullParser.END_TAG) {
		String tag = xpp.getName();
		if ("Folder".equals(tag)) {
		    currf = currf.m_parent;
		}
	    } else if(eventType == XmlPullParser.TEXT) {
		// ... 
	    }
	    eventType = xpp.next();
	}
	} catch (Exception ex) {
	    System.out.println("problem loading file "+ex);
	    ex.printStackTrace();
	}
    }


    void interp() {
	m_root.interp();
	repaint();
    }

    void go() {
	System.out.println("Animating...");
	m_root.resetAnim(true); // activate 1st

	repaint();
	//	try { Thread.sleep(300); } catch (Exception ex) {}
	if (m_anim != null) { m_anim.stopAnim(); }
	m_anim = new Anim(this, true);
	m_anim.start();
    }

    int m_currT;
    
    // returns num alive
    int step(int t) {
	return move();
    }
    
    int numPlaces() { return m_root.countThings(); }

    static Random m_rand = new Random();

   
    // 0  1  2  3  4  5
    int move() {
	// *** m_force = false;
	//	System.out.println("stopping notes...");
	Slime.stopNotes(); // from previous time...
	//	System.out.println(" ... stopped notes.");	
	int alive = m_root.spreadActivity();
	//System.out.println(" ... spread activity.");	
	m_root.spreadFood();
	//System.out.println(" ... spread food.");	
	m_root.moveActivity();

	//	System.out.println(" ... move activity.");	
	m_root.nextState();

	m_root.checkReset();

	if (alive==0) m_force = true;
	//System.out.println(" ... returning "+alive);	
	return alive;
    }

    Color[] m_cols = null;
    int NUMCOLS = 5;
    Color getCol(int i) {
	if (m_cols == null) {
	    m_cols = new Color[NUMCOLS];
	    for (int j=0; j<NUMCOLS; j++) {
		m_cols[j] = Color.getHSBColor( 0.45f, 0.95f, 0.5f + (float) (0.5f*(NUMCOLS-1-j)/NUMCOLS) );
	    }	    
	}
	if (i<0 || i>=NUMCOLS) {
	    return Color.black;
	}
	if (i==0) { return Color.green; }
	return m_cols[i];
    }

    int getW(int w) {
	return (int) ((w) * m_zoom);
    }

    // get distance between 2 points.
    int getD(int p1, int p2) {
	return getX(p2) - getX(p1);
	//	return (int) ((w) * m_zoom);
    }

    // get distance between 2 points : start, width
    int getE(int p1, int w) {
	return getX(p1+w) - getX(p1);
    }

    int getX(int x) {
	return (int)((x) * m_zoom);
    }

    int getY(int y) {
	return (int)((y) * m_zoom);
    }

    // translate pix to real coords
    int pixToX(int x) {
	return (int)(x / m_zoom);
    }

    int pixToY(int y) {
	return (int)(y / m_zoom);
    }

    static Color m_bg = new Color(120,120,140);
    

    // Paint contents of current folder.
    public void paint(Graphics gg) {   

	// Apply scale + zoom...
	int w = getWidth();
	int h = getHeight();
	
	if (m_force) {
	    gg.setColor(m_bg);
	    gg.fillRect(0,0,w,h);
	}

	Graphics2D g2d = (Graphics2D)gg;

	g2d.setRenderingHint( RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON );
	//	g2d.setRenderingHint( RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON );

	m_root.paintChildren(gg, this, true); //  m_force);

	if (m_drawBox) {
	    gg.setColor(Color.black);
	    gg.drawRect(getX(m_boxx), getY(m_boxy), getW(m_boxx2-m_boxx), getW(m_boxy2-m_boxy));
	}
    }
}



class Anim extends Thread {    
    MapPanel m_panel;
    boolean m_fwd = true;

    int speed = 10;

    static boolean m_done = false;

    Runnable m_sync = new Runnable() {
	    public void run() {
	    }
	};

    Anim(MapPanel panel, boolean fwd) {
	m_panel = panel;
	m_fwd = fwd;
    }

    void swingsync() throws InterruptedException {
	try {
	    boolean oldreally;
	    SwingUtilities.invokeAndWait(m_sync);
	    oldreally = m_panel.m_reallyForce;
	    // *** m_panel.m_force = m_panel.m_reallyForce;
	    m_panel.repaint();
	    SwingUtilities.invokeAndWait(m_sync);   
	    if (oldreally) m_panel.m_reallyForce = false;
	    m_panel.m_force = true;
	} catch (Exception ex) {
	}
    }

    void stopAnim() {
	m_done = true;
	try {Thread.sleep(10);} catch (Exception e) {} // should wait on semaphore instead.
    }

    public void run() {
	try {
	    //   m_panel.repaint();
	    //    Thread.sleep(100);
	    m_done = false;

	    System.out.println("Animating thr started");
	    int np = m_panel.numPlaces() * 10;
	    if (m_fwd) {
		for (int i=0; i<np && !m_done; i++) {
		    long t1 = System.currentTimeMillis();
		    // System.out.println("Animating step "+i);
		    final int step = i;

		    SwingUtilities.invokeAndWait( new Runnable() {
			    public void run() {
				// System.out.println("invoking step on swing tread...");
				
				if (m_panel.step(step) == 0) {
				    m_done = true;		    
				}
								
				// System.out.println("... step done");
			    }
			});
		    
		    //		    m_panel.repaint();
		    /*		    
		    if (m_panel.step(i) == 0) {
			System.out.println("activity died out.");
			swingsync();
			break;
		    }
		    */
		    swingsync();
		    Thread.sleep(10);
		    long elapsed = System.currentTimeMillis() - t1;
		    if (elapsed < speed) {
			Thread.sleep(speed - elapsed);
		    }
		    speed = m_panel.m_slime.m_delay; // update from main panel.
		}
	    }
	    else {
		for (int i=np; i>0; i--) {
		    m_panel.step(i);
		    m_panel.repaint();
		    Thread.sleep(100);
		}
	    }
	    System.out.println("Animating thr finished");

	} catch (Exception ex) {
	    ex.printStackTrace();
	    System.out.println("Problem anim "+ex);
	}
    }
}


public class Slime {
    MapPanel m_panel;
    MidiSynth m_synth;
    static Slime m_inst;
    JScrollPane m_scroll;

    KeyboardReader m_keyboardReader = null;

    public Slime() {
	m_inst = this;
	m_synth = new MidiSynth();
	m_synth.open();
	try {
	    // From file
	    //Sequence sequence = MidiSystem.getSequence(new File("jamesbond.mid"));    
	    //Sequence sequence = MidiSystem.getSequence(new File("bohemian.mid"));    
	    Sequence sequence = MidiSystem.getSequence(new File("fugued.mid"));    
	    //Sequence sequence = MidiSystem.getSequence(new File("odetojoy.mid"));    
	    FileWriter out = new FileWriter(new File("out.xml"));

	    HashMap oldplace = new HashMap();
	    out.write("<Folder id='0'> \n");
	    // out.write("  <Place x='10' y='50'/>\n"); // start place...

	    // Create a sequencer for the sequence
	    Sequencer sequencer = MidiSystem.getSequencer();
	    sequencer.open();
	    sequencer.setSequence(sequence);
	    
	    int maxnotes = 400;
	    int note = 1;

	    // get the tracks..
	    int lasttime = 1; // index of prev time place.
	    int nextplace = 2; // index of next place to add
	    long lastTick = 0;
	    int lastxp =0;
	    Track[] tracks = sequence.getTracks();
	    StringWriter outs = new StringWriter();

	    int xp1 = -1; // first xp...

	    int res = 10;

	    for (int t=0; t<tracks.length; t++) { // was tracks.length
		int evts = tracks[t].size();
		int n1 = note;
		for (int e=0; e<evts && note < maxnotes; e++) {
		    MidiEvent me = tracks[t].get(e);

		    
		    // print it out...
		    long tick = me.getTick();
		    int xp = (int) tick;
		    // xp -= tick % res;
		    
		    if (xp1==-1) {
			out.write("  <Place x='10' y='50' id='1'/>\n"); // start place...
			xp1 = xp;
		    }
		    else if (xp > lastxp) {
			while (lastTick < tick) {
			    lastTick += res;
			    xp = (int) lastTick;

			    out.write("  <Place x='"+(xp-xp1+10)+"' y='100' id='"+nextplace+"'/>\n");			    
			    out.write("  <Trans from='"+lasttime+"' to='"+nextplace+"'/>\n");
			    oldplace.put(lastTick, nextplace);
			    lasttime = nextplace;
			    nextplace ++;

			    lastxp = xp;
			}
		    }
		    
		    MidiMessage mm = me.getMessage();
		    if (mm instanceof ShortMessage) {
			ShortMessage sm = (ShortMessage)mm;
			int cmd = sm.getCommand();
			int d1 = sm.getData1();
			int d2 = sm.getData2();
			if (cmd == ShortMessage.NOTE_ON && d2 != 0) {
			    int yp = 700 - d1 * 6;
			    out.write("  <Place x='"+(xp-xp1+10)+"' y='"+yp+"' id='"+nextplace+"' output='Note'/>\n");
			    
			    out.write("  <Trans from='"+lasttime+"' to='"+nextplace+"' onset='true'/>\n");
			    nextplace ++;
			    note++;
			}
		    }
		}			
	    }


	    out.write("</Folder>\n");
	    out.close();

	    // Start playing
	    //sequencer.start();
	} catch (Exception e) {
	}

    }

    // play a note...
    static void playNote(int inst, int n) {
	n = 1000-n;
	n /= 6;
	n %= 200;
        System.out.println("Playing a note..."+inst+" "+n);
	m_inst.m_synth.play(2, n , 120);
	// m_inst.m_synth.click(); // play(inst, n, 120);
	try { Thread.sleep(100); } catch (Exception e) {}
	m_inst.m_synth.release();
    }
    
    static void startNote(int n) {
	n = 500-n;
	n /= 6;
	n %= 200;
	//	System.out.println("Starting a note "+n);
	m_inst.m_synth.play(2, n, 120);
	//System.out.println("Told synth to start it "+n);
    }

    static void stopNotes() {
	m_inst.m_synth.release();
    }

    
    JPanel m_placeDetails;
    JTextField m_placeName;
    JCheckBox m_isStatePlace;
    JCheckBox m_initActive;
    JCheckBox m_noteActive;
    JComboBox m_outputCmd;
    JButton m_apply;
    JSlider m_speed;
    int m_delay = 100;

    Thing m_selThing; 

    String[] cmds = {"(none)", "Recorder", "MidiReader", "MidiPlayer", "TextFileReader", "KeyboardReader", "TextWriter", "MouseReader", "Drawer"};

    int getSelIndex(String cname) {
	for (int i=0; i<cmds.length; i++) {
	    if (cmds[i].equals(cname)) { return i; }
	}
	System.out.println("No entry found for "+cname);
	return 0;
    }

    void updatePlacePanel(Thing thing) {

	if (m_selThing != null) {
	    updatePlaceFromPanel();
	}

	if (thing instanceof Folder) {
	    m_placeName.setEnabled(true);
	    m_apply.setEnabled(true);	    
	    m_isStatePlace.setEnabled(false);	    
	    m_initActive.setEnabled(false);	 
   	    m_noteActive.setEnabled(false);
	    m_outputCmd.setEnabled(true);
	    Folder f = (Folder)thing;
	    if (f.m_behaviour != null) {		
		
		m_outputCmd.setSelectedIndex(getSelIndex(f.m_behaviour.getClass().getName()));
		// *** Adapt for alt behavuours ***
	    }
	    else {
		m_outputCmd.setSelectedIndex(0);
	    }
	    m_placeName.setText(thing.m_label);
	    m_selThing = thing;
	} 
	else if (thing instanceof Place) {
	    m_placeName.setEnabled(true);
	    m_apply.setEnabled(false);	    
	    m_isStatePlace.setEnabled(true);	    
	    m_initActive.setEnabled(true);	    	    
	    m_noteActive.setEnabled(true);	    	    
	    m_outputCmd.setEnabled(false);	    	       
	    
	    m_placeName.setText(thing.m_label);
	    m_selThing = thing;

	    Place place = (Place) thing;

	    m_isStatePlace.setSelected(place.isState);
	    m_initActive.setSelected(place.initDistFromActive == 0);
	    
	    if (place.m_outputCmd==null) {
		//	m_outputCmd.setSelectedIndex(0);
		m_noteActive.setSelected(false);
	    }
	    else { 
		//	m_outputCmd.setSelectedItem(place.m_outputCmd);
		m_noteActive.setSelected("Note".equals(place.m_outputCmd));
	    }	    
	    m_apply.setEnabled(true); // while the place is valid... needs check for deletion...
	}
	else if (thing==null) {
	    m_placeName.setEnabled(false);
	    m_apply.setEnabled(false);	    
	    m_isStatePlace.setEnabled(false);	    
	    m_initActive.setEnabled(false);	    	    
	    m_noteActive.setEnabled(false);	    	    
	    m_outputCmd.setEnabled(false);	    
	    m_selThing = null;
	}
    }


    void updatePlaceFromPanel() {
	if (m_selThing==null) return;
	String txt = m_placeName.getText();
	if ("".equals(txt)) {
	    txt = null; 
	}

	// for all selected places...
	LinkedList<Thing> sel = m_panel.getSelectedThings();
	Iterator<Thing>iter = sel.iterator();

	while (iter.hasNext()) {
	    Thing t = iter.next();
	    if (t instanceof Place && m_selThing instanceof Place) {
		t.m_label = txt;
		Place p = (Place)t;
		p.isState = m_isStatePlace.isSelected();
		p.initDistFromActive = m_initActive.isSelected() ? 0 : -1;
		if (m_noteActive.isSelected()) {
		    p.m_outputCmd = "Note";
		    // overrides pulldown.
		}
		else {
		    p.m_outputCmd = null; // m_outputCmd.getSelectedIndex()==0 ? null : m_outputCmd.getSelectedItem().toString();
		}
	    }
	    else if (t instanceof Folder && m_selThing instanceof Folder) {
		t.m_label = txt;
		
		Folder f = (Folder)t;
		String item = m_outputCmd.getSelectedItem().toString();
		if ( item.equals("Recorder")) {
		    if (!(f.m_behaviour instanceof Recorder)) {
			f.setBehaviour(new Recorder(f));
		    }
		}
		else if ( item.equals("KeyboardReader")) {
		    if (!(f.m_behaviour instanceof KeyboardReader)) {
			KeyboardReader reader = new KeyboardReader(f, m_panel);
			f.setBehaviour(reader);
			m_keyboardReader = reader;
		    }
		}
		else {
		    f.setBehaviour(null);
		}
	    }
	}
    }

    void doit() {
	m_inst = this;
	JFrame frame = new JFrame("Slime trails");

	//	frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
	frame.addWindowListener(new WindowAdapter() {
                public void windowClosing(WindowEvent e) {
                    // clean up ...
                    System.out.println("Cleaning up synth nicely...");
		    if (m_synth != null) {
			m_synth.release();
			m_synth.close();
		    } 
                    System.exit(0);
                }
            });

	m_panel = new MapPanel(this);	

	m_panel.addKeyListener(new KeyboardAdapter(this));

	m_scroll = new JScrollPane(m_panel, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
	

	JPanel m_stateDetails = new JPanel(new FlowLayout());

	m_placeName = new JTextField( 20);
	//	m_placeName.setPreferredSize(new Dimension(100,10));
	m_isStatePlace = new JCheckBox("state");
	m_initActive = new JCheckBox("init");
	m_noteActive = new JCheckBox("note");
	m_outputCmd = new JComboBox(cmds);

	m_apply = new JButton("Apply");
	m_apply.setEnabled(false);


	m_speed = new JSlider(JSlider.HORIZONTAL, 1, 200, 5);
	m_speed.setMajorTickSpacing(50);
	m_speed.setPaintTicks(true);
	m_stateDetails.add(m_placeName);
	m_stateDetails.add(m_isStatePlace);
	m_stateDetails.add(m_initActive);
	m_stateDetails.add(m_noteActive);
	m_stateDetails.add(m_outputCmd);
	m_stateDetails.add(m_apply);
	m_stateDetails.add(m_speed);


	m_speed.addChangeListener(new ChangeListener() {
		public void stateChanged(ChangeEvent e) {
		    JSlider source = (JSlider)e.getSource();
		    int fps = (int)source.getValue();			
		    m_delay = 1000 / fps;			    			    
		}
	    });
	AbstractAction upd = new AbstractAction() {
		public void actionPerformed(ActionEvent ae) {
		    updatePlaceFromPanel();
		    m_panel.repaint();
		}
	    };
	m_apply.addActionListener( upd);
	m_isStatePlace.addActionListener( upd);
	m_placeName.addActionListener( new AbstractAction() {
		public void actionPerformed(ActionEvent ae) {
		    // m_outputCmd.setSelectedItem(cmds[2]);
		    updatePlaceFromPanel();
		    // m_selPlace.m_outputCmd = "Text";
		    m_panel.repaint();
		}
	    }
				       );
	    
	m_initActive.addActionListener(upd);
	m_noteActive.addActionListener(upd);
       
	JPanel main = new JPanel(new BorderLayout());
	main.add(m_stateDetails, BorderLayout.NORTH);

	JPanel buts = new JPanel();

	JButton reset = new JButton("Reset");
	JButton interp = new JButton("Interp");
	JButton go = new JButton("Go");
	//	JButton back = new JButton("Back");
	JButton delete = new JButton("Delete");
	JButton group = new JButton("Group");
	JButton ungroup = new JButton("Ungroup");
	JButton save = new JButton("Save");
	JButton load = new JButton("Load");

	// parameters of selected item...
	

	buts.add(reset);
	reset.addActionListener( new AbstractAction() {
		public void actionPerformed(ActionEvent ae) {
		    m_panel.reset();
		}
	    });

	buts.add(interp);
	interp.addActionListener( new AbstractAction() {
		public void actionPerformed(ActionEvent ae) {
		    m_panel.interp();
		}
	    });

	buts.add(go);
	go.addActionListener( new AbstractAction() {
		public void actionPerformed(ActionEvent ae) {
		    m_panel.go();
		}
	    });

	/*
	buts.add(back);
	back.addActionListener( new AbstractAction() {
		public void actionPerformed(ActionEvent ae) {
		    m_panel.back();
		}
	    });
	*/
	buts.add(save);
	save.addActionListener( new AbstractAction() {
		public void actionPerformed(ActionEvent ae) {
		    m_panel.save();
		}
	    });
	buts.add(load);
	load.addActionListener( new AbstractAction() {
		public void actionPerformed(ActionEvent ae) {
		    m_panel.load();
		}
	    });

	buts.add(delete);
	delete.addActionListener( new AbstractAction() {
		public void actionPerformed(ActionEvent ae) {
		    m_panel.deleteSelected();
		}
	    });

	buts.add(group);
	group.addActionListener( new AbstractAction() {
		public void actionPerformed(ActionEvent ae) {
		    m_panel.groupSelected();
		}
	    });
	buts.add(ungroup);
	ungroup.addActionListener( new AbstractAction() {
		public void actionPerformed(ActionEvent ae) {
		    m_panel.ungroupSelected();
		}
	    });

	frame.getContentPane().add(main);
	main.add(m_scroll);
	main.add(buts, BorderLayout.SOUTH);
	frame.pack();
	frame.show();
    }


    public static void main(String[] args) {
	final Slime s = new Slime();
	try {SwingUtilities.invokeAndWait(new Runnable() {
		public void run() {
		    s.doit();
		}
	    });
	} catch (Exception ex) {
	    System.out.println(ex);
	    ex.printStackTrace();
	    
	}
    }

}
